#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

from plenopticam.lfp_extractor import LfpViewpoints
from plenopticam.misc import PlenopticamError


class LfpRearranger(LfpViewpoints):

    def __init__(self, lfp_img_align=None, *args, **kwargs):
        super(LfpRearranger, self).__init__(*args, **kwargs)

        self._lfp_img_align = lfp_img_align if lfp_img_align is not None else None
        self._dtype = self._lfp_img_align.dtype if self._lfp_img_align is not None else self._vp_img_arr.dtype

    def _init_vp_img_arr(self):
        """ initialize viewpoint output image array """

        if len(self._lfp_img_align.shape) == 3:
            m, n, p = self._lfp_img_align.shape
        elif len(self._lfp_img_align.shape) == 2:
            m, n, p = self._lfp_img_align.shape[:2] + (1,)
        else:
            raise PlenopticamError('Dimensions %s of provided light-field not supported', self._lfp_img_align.shape,
                                   cfg=self.cfg, sta=self.sta)

        self._vp_img_arr = np.zeros([int(self._size_pitch), int(self._size_pitch),
                                     int(m/self._size_pitch), int(n/self._size_pitch), p], dtype=self._dtype)

    def _init_lfp_img_align(self):
        """ initialize micro image output image array """

        if len(self._vp_img_arr.shape) == 5:
            m, n, p = self._vp_img_arr.shape[2:]
        elif len(self._vp_img_arr.shape) == 4:
            m, n, p = self._vp_img_arr.shape[2:] + (1,)
        else:
            raise PlenopticamError('Dimensions %s of provided light-field not supported', self._vp_img_arr.shape,
                                   cfg=self.cfg, sta=self.sta)

        # create empty array
        m *= self._vp_img_arr.shape[0]
        n *= self._vp_img_arr.shape[1]
        self._lfp_img_align = np.zeros([m, n, p], dtype=self._dtype)

        # update angular resolution parameter
        self._size_pitch = self._vp_img_arr.shape[0] if self._vp_img_arr.shape[0] == self._vp_img_arr.shape[1] else float('inf')

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # rearrange light-field to viewpoint representation
        self.compose_viewpoints()

    def compose_viewpoints(self):
        """
        Conversion from aligned micro image array to viewpoint array representation. The fundamentals behind the
        4-D light-field transfer were derived by Levoy and Hanrahans in their paper 'Light Field Rendering' in Fig. 6.
        """

        # print status
        self.sta.status_msg('Viewpoint composition', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # initialize basic light-field parameters
        self._init_vp_img_arr()

        # rearrange light field to multi-view image representation
        for j in range(self._size_pitch):
            for i in range(self._size_pitch):

                # check interrupt status
                if self.sta.interrupt:
                    return False

                # extract viewpoint by pixel rearrangement
                self._vp_img_arr[j, i, ...] = self._lfp_img_align[j::self._size_pitch, i::self._size_pitch, :]

                # print status
                percentage = (j * self._size_pitch + i + 1) / self._size_pitch ** 2
                self.sta.progress(percentage*100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def decompose_viewpoints(self):
        """
        Conversion from viewpoint image array to aligned micro image array representation. The fundamentals behind the
        4-D light-field transfer were derived by Levoy and Hanrahans in their paper 'Light Field Rendering' in Fig. 6.
        """

        # print status
        self.sta.status_msg('Viewpoint decomposition', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # initialize basic light-field parameters
        self._init_lfp_img_align()

        # rearrange light field to multi-view image representation
        for j in range(self._size_pitch):
            for i in range(self._size_pitch):

                # check interrupt status
                if self.sta.interrupt:
                    return False

                # extract viewpoint by pixel rearrangement
                self._lfp_img_align[j::self._size_pitch, i::self._size_pitch, :] = self._vp_img_arr[j, i, :, :, :]

                # print status
                percentage = (j * self._size_pitch + i + 1) / self._size_pitch ** 2
                self.sta.progress(percentage*100, self.cfg.params[self.cfg.opt_prnt])

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import getopt
import sys, os

# local imports
from plenopticam import lfp_reader
from plenopticam import lfp_calibrator
from plenopticam import lfp_aligner
from plenopticam import lfp_extractor
from plenopticam import lfp_refocuser
from plenopticam.lfp_reader.top_level import SUPP_FILE_EXT
from plenopticam import misc
from plenopticam.cfg import PlenopticamConfig
from plenopticam.cfg.constants import CALI_METH, SMPL_METH, CLIF_OPTS, CLIF_SHRT
from plenopticam import __version__
from plenopticam.gui.top_level import PlenopticamApp


def usage():
    """ prints help message """

    print("Usage: plenopticam <options>\n")
    print("Options:")
    print("-g                                Open up graphical user interface")
    print("-f <filepath>, --file=<filepath>  Specify image file or folder to process")
    print("-c <calipath>, --cali=<calipath>  Specify calibration file to process")
    print("-p <number>,   --patch=<number>   Patch size")
    print("-r <list>,     --refo=[0, 2]      Refocusing range")
    print("-m <str>,      --meth='area'      Calibration method, e.g.:")
    print("                                  "+', '.join(['"'+m+'"' for m in CALI_METH]))
    print("-s <str>,      --smpl='global'    Resampling method, e.g.:")
    print("                                  "+', '.join(['"'+m+'"' for m in SMPL_METH]))
    print("-h,            --help             Print this help message")
    print("")
    # boolean options
    print("--refi                            Refocusing refinement")
    print("--pflu                            Scheimpflug refocus")
    print("--vign                            De-vignetting")
    print("--cont                            Contrast automation")
    print("--colo                            Color Equalization")
    print("--awb                             Auto white balance")
    print("--sat                             Saturation automation")
    print("--lier                            Hot pixel treatment")
    print("--arti                            Artifact removal")
    print("--remo                            Override output folder")
    print("")


def parse_options(argv, cfg):
    """ set config dictionary according to command line user arguments """

    try:
        opts, args = getopt.getopt(argv, CLIF_SHRT, CLIF_OPTS)
    except getopt.GetoptError as e:
        print(e)
        sys.exit(2)

    if opts:
        for (opt, arg) in opts:
            if opt in ("-g", "--gui"):
                PlenopticamApp().mainloop()
                sys.exit()
            if opt in ("-h", "--help"):
                usage()
                sys.exit()
            if opt in ("-f", "--file"):
                cfg.params[cfg.lfp_path] = arg.strip(" \"\'")
            if opt in ("-c", "--cali"):
                cfg.params[cfg.cal_path] = arg.strip(" \"\'")
            if opt == "--meta":
                cfg.params[cfg.cal_meta] = arg.strip(" \"\'") if arg.strip(" \"\'") else ''
            if opt in ("-m", "--meth"):
                cfg.params[cfg.cal_meth] = arg.strip(" \"\'") if arg.strip(" \"\'") in CALI_METH else CALI_METH[0]
            if opt in ("-s", "--smpl"):
                cfg.params[cfg.smp_meth] = arg.strip(" \"\'") if arg.strip(" \"\'") in SMPL_METH else SMPL_METH[0]
            if opt in ("-p", "--patch"):
                cfg.params[cfg.ptc_leng] = misc.str2type(arg)
            if opt in ("-r", "--refo"):
                refo_range = misc.str2list(arg)
                cfg.params[cfg.ran_refo] = refo_range if isinstance(refo_range, list) else [0, 2]
            if opt == "--rcal":
                cfg.params[cfg.opt_cali] = True
            if opt == "--vign":
                cfg.params[cfg.opt_vign] = True
            if opt == "--lier":
                cfg.params[cfg.opt_lier] = True
            if opt == "--cont":
                cfg.params[cfg.opt_cont] = True
            if opt == "--colo":
                cfg.params[cfg.opt_colo] = True
            if opt == "--awb":
                cfg.params[cfg.opt_awb_] = True
            if opt == "--sat":
                cfg.params[cfg.opt_sat_] = True
            if opt == "--refi":
                cfg.params[cfg.opt_refi] = True
            if opt == "--pflu":
                cfg.params[cfg.opt_pflu] = True
            if opt == "--arti":
                cfg.params[cfg.opt_arti] = True
            if opt == "--rota":
                cfg.params[cfg.opt_rota] = True
            if opt == "--dbug":
                cfg.params[cfg.opt_dbug] = True
            if opt == "--prnt":
                cfg.params[cfg.opt_prnt] = True
            if opt == "--dpth":
                cfg.params[cfg.opt_dpth] = True
            if opt == "--remo":
                cfg.params[cfg.dir_remo] = True

    return cfg


def main():

    # program info
    print("\nPlenoptiCam v%s \n" % __version__)

    # create config object
    cfg = PlenopticamConfig()
    cfg.default_values()

    # parse options
    cfg = parse_options(sys.argv[1:], cfg)

    # instantiate status object
    sta = misc.PlenopticamStatus()

    # force relative paths to be absolute
    cfg.params[cfg.lfp_path] = os.path.abspath(cfg.params[cfg.lfp_path])
    cfg.params[cfg.cal_path] = os.path.abspath(cfg.params[cfg.cal_path])

    # collect light field image file name(s) based on provided path
    if os.path.isdir(cfg.params[cfg.lfp_path]):
        lfp_filenames = [f for f in os.listdir(cfg.params[cfg.lfp_path]) if f.lower().endswith(SUPP_FILE_EXT)]
        cfg.params[cfg.lfp_path] = os.path.join(cfg.params[cfg.lfp_path], 'dummy.ext')
    elif not os.path.isfile(cfg.params[cfg.lfp_path]):
        lfp_filenames = [misc.select_file(cfg.params[cfg.lfp_path], 'Select plenoptic image')]
    else:
        lfp_filenames = [cfg.params[cfg.lfp_path]]

    if not cfg.params[cfg.cal_path]:
        # manual calibration data selection
        sta.status_msg('\r Please select white image calibration source manually', cfg.params[cfg.opt_prnt])
        # open selection window (at current lfp file directory) to set calibration folder path
        cfg.params[cfg.cal_path] = misc.select_file(cfg.params[cfg.lfp_path], 'Select calibration image')

    # provide number of found images to user
    print("\n %s Image(s) found" % len(lfp_filenames))

    # cancel if file paths not provided
    sta.validate(checklist=lfp_filenames+[cfg.params[cfg.lfp_path]], msg='Canceled due to missing image file path')

    # iterate through light field image(s)
    for lfp_filename in sorted(lfp_filenames):

        # change path to next filename
        cfg.params[cfg.lfp_path] = os.path.join(os.path.dirname(cfg.params[cfg.lfp_path]), lfp_filename)
        print(cfg.params[cfg.lfp_path])
        sta.status_msg(msg='Process file '+lfp_filename, opt=cfg.params[cfg.opt_prnt])

        # remove output folder if option is set
        misc.rmdir_p(cfg.exp_path) if cfg.params[cfg.dir_remo] else None

        try:
            # decode light field image
            aligner = lfp_reader.LfpReader(cfg, sta)
            aligner.main()
            lfp_img = aligner.lfp_img
        except Exception as e:
            misc.PlenopticamError(e, cfg=cfg, sta=sta)
            continue
        # create output data folder
        misc.mkdir_p(cfg.exp_path, cfg.params[cfg.opt_prnt])

        if cfg.cond_auto_find():
            # automatic calibration data selection
            extractor = lfp_calibrator.CaliFinder(cfg, sta)
            extractor.main()
            wht_img = extractor.wht_bay
        else:
            # load white image calibration file
            wht_img = misc.load_img_file(cfg.params[cfg.cal_path])
            # save settings configuration
            cfg.save_params()

        # perform calibration if previously computed calibration data does not exist
        meta_cond = not (os.path.exists(cfg.params[cfg.cal_meta]) and cfg.params[cfg.cal_meta].lower().endswith('json'))
        if meta_cond or cfg.params[cfg.opt_cali]:
            # perform centroid calibration
            calibrator = lfp_calibrator.LfpCalibrator(wht_img, cfg, sta)
            calibrator.main()
            cfg = calibrator.cfg

        # load calibration data
        cfg.load_cal_data()

        #  check if light field alignment has been done before
        if cfg.cond_lfp_align():
            # align light field
            aligner = lfp_aligner.LfpAligner(lfp_img, cfg, sta, wht_img)
            aligner.main()
            lfp_img_align = aligner.lfp_img
        else:
            lfp_img_align = None

        # extract viewpoint data
        lfp_calibrator.CaliFinder(cfg).main()
        extractor = lfp_extractor.LfpExtractor(lfp_img_align, cfg=cfg, sta=sta)
        extractor.main()

        # do refocusing
        if cfg.params[cfg.opt_refo]:
            extractor = lfp_refocuser.LfpRefocuser(extractor.vp_img_linear, cfg=cfg, sta=sta)
            extractor.main()

    return True


if __name__ == "__main__":

    sys.exit(main())



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam import misc
from plenopticam.cfg import constants as c

# external libs
import numpy as np
import os


class LfpScheimpflug(object):

    def __init__(self, refo_stack=None, lfp_img=None, cfg=None, sta=None):

        self.refo_stack = refo_stack
        self.lfp_img = lfp_img
        self.cfg = cfg
        self.sta = sta

        self.fp = os.path.join(os.path.splitext(self.cfg.params[self.cfg.lfp_path])[0], 'scheimpflug')

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Scheimpflug extraction \n', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # create output folder
        misc.mkdir_p(self.fp)

        # iterate through directions
        for direction in list(c.PFLU_VALS):
            self.cfg.params[self.cfg.opt_pflu] = direction
            self.sta.status_msg(direction+':')

            if self.refo_stack is not None:
                self.scheimpflug_from_stack()
            elif self.lfp_img is not None:
                self.scheimpflug_from_scratch()

        return True

    def scheimpflug_from_stack(self):

        a_start, a_stop = (0, len(self.refo_stack))

        if len(self.refo_stack[0].shape) == 3:
            m, n, p = self.refo_stack[0].shape
        else:
            m, n, p = (self.refo_stack[0].shape[0], self.refo_stack[0].shape[1], 1)

        scheimpflug_img = np.zeros([m, n, p])

        # map generation
        a_x = np.linspace(0, a_stop-a_start, n+2, dtype='int')[1:-1]      # flooring via integer type while excluding
        a_y = np.linspace(0, a_stop-a_start, m+2, dtype='int')[1:-1]
        a_map_x = np.outer(np.ones(m, dtype='int'), a_x)
        a_map_y = np.outer(a_y, np.ones(n, dtype='int'))

        # vertical orientation (default)
        a_map = a_map_y
        # horizontal orientation
        if self.cfg.params[self.cfg.opt_pflu] == c.PFLU_VALS[1]:
            a_map = a_map_x
        # diagonal orientation
        elif self.cfg.params[self.cfg.opt_pflu] == (c.PFLU_VALS[2] or c.PFLU_VALS[3]):
            # swap refocusing directions if option set
            if self.cfg.params[self.cfg.opt_pflu] == c.PFLU_VALS[3]:
                a_map_x, a_map_y = a_map_x[::-1], a_map_y[::-1]
            a_map = np.mean(np.stack([a_map_x, a_map_y]), dtype='int', axis=0)

        for y in range(m):
            for x in range(n):
                scheimpflug_img[y, x] = self.refo_stack[a_map[y, x]][y, x]

                # check interrupt status
                if self.sta.interrupt:
                    return False

                # print status
                percentage = (((y*n+x+1)/(m*n))*100)
                self.sta.progress(percentage, self.cfg.params[self.cfg.opt_prnt])

        # write image file to hard drive
        a_ran = self.cfg.params[self.cfg.ran_refo]
        fn = 'scheimpflug_' + str(a_ran[0]) + '_' + str(a_ran[-1]) + '_' + self.cfg.params[self.cfg.opt_pflu] + '.png'
        misc.save_img_file(misc.Normalizer(scheimpflug_img).uint16_norm(), os.path.join(self.fp, fn))

        return True

    def scheimpflug_from_scratch(self):     # pragma: no cover

        patch_len = self.cfg.params[self.cfg.ptc_leng]
        a_start, a_stop = self.cfg.params[self.cfg.ran_refo]

        img = self.lfp_img.astype('float') / patch_len
        m, n, P = self.lfp_img.shape
        overlap = int(abs(a_stop) * (patch_len - 1))
        #tilt_vec_y = np.linspace(a_start, planes, m).astype('uint')
        a_x = np.linspace(a_start, a_stop, n + overlap).astype('int') # flooring values instead of rounding => 1st round, then int()
        a_y = np.zeros(m).astype('int')#np.linspace(a_start, planes, m).astype('uint')
        a_xy = np.ones([m, n + overlap]).astype('int') * a_x

        # initialize matrices for intermediate and plane results
        hor_refo = np.zeros([m, n + overlap, P], dtype='float')
        ver_refo = np.zeros([m + overlap, n + overlap, P], dtype='float')
        fraction_vec = np.zeros([patch_len, P], dtype='float')

        # horizontal scheimpflug shift and integration
        for y in range(m):
            for x in range(n):

                # prevent from taking adjacent pixel being beyond image border
                if x + patch_len < n:
                    adj_idx = x + patch_len
                else:
                    adj_idx = x

                for p in range(P):
                    fraction_vec[:, p] = np.linspace(img[y, x, p], img[y, adj_idx, p], patch_len)

                # load refocus value at x,y
                a = a_xy[y, x]

                # consider negative shift
                negative_a = int((patch_len - 1) * abs(a)) if a < 0 else 0

                # centralize row
                row_shift = int((a_y.max() - a)*(patch_len - 1)/2)

                newX = int(x + np.mod(x, patch_len) * (a - 1) + negative_a) + row_shift
                hor_refo[y, newX:newX + patch_len, :] = hor_refo[y, newX:newX + patch_len, :] + fraction_vec

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status
            self.sta.progress((y+1)/m*50, self.cfg.params[self.cfg.opt_prnt])

        # vertical scheimpflug shift and integration
        new_n = n + int(abs(a_x.max()) * (patch_len - 1))
        for x in range(new_n):
            for y in range(m):

                if y + patch_len < m:
                    adj_idx = y + patch_len
                else:
                    adj_idx = y

                frac_vec = np.zeros([patch_len, int(n + abs(a_x.max()) * (patch_len - 1)), P], dtype='float')
                for p in range(P):
                    frac_vec[:, x, p] = np.linspace(hor_refo[y, x, p], hor_refo[adj_idx, x, p], patch_len)

                # load refocus value at x,y
                a = a_xy[y, x]

                # consider negative shift
                negative_a = int((patch_len - 1) * abs(a)) if a < 0 else 0

                # centralize column
                column_shift = int((a_x.max()-a)*(patch_len-1)/2)

                # put interpolated vector to refocusing plane
                newY = int(y + np.mod(y, patch_len) * (a - 1) + negative_a) + column_shift
                ver_refo[newY:newY + patch_len, :, :] = ver_refo[newY:newY + patch_len, :, :] + frac_vec

            # print progress status
            self.sta.progress(((x+1)/new_n+.5)*100, self.cfg.params[self.cfg.opt_prnt])

        # write image file to hard drive
        img = misc.Normalizer(ver_refo).uint16_norm()
        misc.save_img_file(img, os.path.join(self.fp, 'scheimpflug_' + str(patch_len) + 'px.tiff'))

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.lfp_extractor import LfpViewpoints
from color_space_converter import yuv_conv

from scipy.signal import medfilt
import numpy as np


class LfpOutliers(LfpViewpoints):

    def __init__(self, *args, **kwargs):
        super(LfpOutliers, self).__init__(*args, **kwargs)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # remove outliers
        self.proc_vp_arr(self.correct_outliers, msg='Pixel outlier removal')

    def correct_outliers(self, img, n=2, perc=.2):

        for j in range(n, img.shape[0]-n):
            for i in range(n, img.shape[1]-n):

                win = img[j-n:j+n+1, i-n:i+n+1]

                # hot pixel detection
                num_hi = len(win[win > img[j, i]*(1-perc)])

                # dead pixel detection
                num_lo = len(win[win < img[j, i]*perc])

                if num_hi < win.size/(2*n+1) or num_lo > win.size/(2*n+1):
                    # replace outlier by average of all directly adjacent pixels
                    img[j, i] = (sum(sum(img[j-1:j+2, i-1:i+2]))-img[j, i])/8.

                # check interrupt status
                if self.sta.interrupt:
                    return False

        return img

    def correct_luma_outliers(self, img, n=2, perc=.2):

        # luma channel conversion
        luma = yuv_conv(img.copy())[..., 0]

        for j in range(n, luma.shape[0]-n):
            for i in range(n, luma.shape[1]-n):
                win = luma[j-n:j+n+1, i-n:i+n+1]

                # hot pixel detection
                num_hi = len(win[win > luma[j, i]*(1-perc)])

                # dead pixel detection
                num_lo = len(win[win < luma[j, i]*perc])

                if num_hi < win.size/(2*n+1) or num_lo < win.size/(2*n+1):
                    # replace outlier by average of all directly adjacent pixels
                    img[j, i, :] = (sum(sum(img[j-1:j+2, i-1:i+2, :]))-img[j, i, :])/8.

                # check interrupt status
                if self.sta.interrupt:
                    return False

        return img

    @staticmethod
    def channel_outliers_filter(channel, perc=.999):

        # create copy of channel for filtering
        arr = channel.copy()

        # perform filter convolution
        filt_img = medfilt(arr, kernel_size=(3, 3))
        #filt_img = median_filter(arr, size=2)

        # compute absolute differences per pixel
        diff_img = abs(arr-filt_img)
        del arr

        # obtain intensity threshold for pixels that have to be replaced
        diff_img /= diff_img.max()
        threshold = np.percentile(diff_img, perc*100)

        # replace pixels above threshold by median filtered pixels while ignoring image borders (due to 3x3 kernel)
        channel[1:-1, 1:-1][diff_img[1:-1, 1:-1] > threshold] = filt_img[1:-1, 1:-1][diff_img[1:-1, 1:-1] > threshold]

        return channel

    def img_outliers_filter(self, img, perc=.999):

        if len(img.shape) == 3:
            for i in range(img.shape[2]):
                img[..., i] = self.channel_outliers_filter(img[..., i], perc=perc)
        elif len(img.shape) == 2:
            img = self.channel_outliers_filter(img)

        return img



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.lfp_extractor.lfp_viewpoints import LfpViewpoints
from plenopticam.lfp_refocuser.lfp_shiftandsum import LfpShiftAndSum
from plenopticam.lfp_refocuser.lfp_scheimpflug import LfpScheimpflug
from plenopticam.lfp_extractor.lfp_exporter import LfpExporter
from plenopticam.lfp_extractor.lfp_contrast import LfpContrast
from plenopticam.misc import GammaConverter


class LfpRefocuser(LfpViewpoints):

    def __init__(self, vp_img_arr=None, *args, **kwargs):
        super(LfpRefocuser, self).__init__(*args, **kwargs)

        # input variables
        self.vp_img_arr = vp_img_arr

        # output variables
        self.refo_stack = []

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # refocused image stack
        if self.cfg.params[self.cfg.opt_refo]:
            self.shift_and_sum()

        # scheimpflug focus
        if self.cfg.params[self.cfg.opt_pflu]:
            self.scheimpflug()

        return True

    def shift_and_sum(self):

        # refocusing
        obj = LfpShiftAndSum(vp_img_arr=self.vp_img_arr, cfg=self.cfg, sta=self.sta)
        obj.main()
        self.refo_stack = obj.refo_stack
        del obj

        # color management automation
        if self.refo_stack is not None and not self.sta.interrupt:
            self.refo_stack = LfpContrast().auto_hist_align(self.refo_stack, ref_img=self.refo_stack[0], opt=True)
            self.refo_stack = GammaConverter().srgb_conv(img=self.refo_stack)

        # write refocused images to hard drive
        if self.refo_stack is not None and not self.sta.interrupt:
            obj = LfpExporter(refo_stack=self.refo_stack, cfg=self.cfg, sta=self.sta)
            obj.export_refo_stack(file_type='png')
            obj.gif_refo()
            del obj

        return True

    def scheimpflug(self):

        obj = LfpScheimpflug(refo_stack=self.refo_stack, cfg=self.cfg, sta=self.sta)
        obj.main()
        del obj

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import PlenopticamStatus
from plenopticam import misc

import numpy as np
import os
from scipy.signal import medfilt


class CfaOutliers(object):

    def __init__(self, *args, **kwargs):

        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else PlenopticamConfig()
        self.sta = kwargs['sta'] if 'sta' in kwargs else PlenopticamStatus()

        self._bay_img = kwargs['bay_img'] if 'bay_img' in kwargs else np.array([])

    def rectify_candidates_bayer(self, bay_img=None, n=2, sig_lev=4):

        # status message
        self.sta.status_msg('Hot pixel detection', self.cfg.params[self.cfg.opt_prnt])

        bay_img = self._bay_img.copy() if bay_img is None else bay_img
        cum_img = np.zeros(bay_img[0::2, 0::2].shape)

        for c in range(4):

            # check interrupt status
            if self.sta.interrupt:
                return False

            # progress update
            percent = c / 4
            self.sta.progress(percent*100, self.cfg.params[self.cfg.opt_prnt])

            # Bayer channel index
            i, j = c//2, c % 2

            # deduct median filtered image
            med_img = medfilt(bay_img[i::2, j::2].copy(), kernel_size=(3, 3))
            m_img = bay_img[i::2, j::2].copy()/bay_img[i::2, j::2].max() - med_img.copy()/np.max(med_img)

            new_img = self.rectify_candidates_channel(channel=bay_img[i::2, j::2].copy(),
                                                      ref_img=m_img, med_img=med_img, n=n, sig_lev=sig_lev+2)

            diff_img = bay_img[i::2, j::2].copy() - new_img
            cum_img[diff_img != 0] = 1

            bay_img[i::2, j::2] = new_img

        # export hot-pixel map
        if self.cfg.params[self.cfg.opt_dbug]:
            hotp_num = len(cum_img[cum_img != 0])
            misc.save_img_file(cum_img, file_path=os.path.join(self.cfg.exp_path, 'hot-pixel_map_'+str(hotp_num)+'.png'))

        # progress update
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        self._bay_img = bay_img

        return True

    def rectify_candidates_channel(self, channel, ref_img=None, med_img=None, n=2, sig_lev=4):

        ref_img = channel if ref_img is None else ref_img
        med_img = channel if med_img is None else med_img

        # pre-select outlier candidates (narrow-down search area to speed-up the process)
        m_val = np.mean(ref_img)
        s_val = np.std(ref_img)
        candidate_idxs = np.where(ref_img > m_val + s_val * sig_lev)

        ref_img = np.zeros_like(channel)
        ref_img[candidate_idxs[0], candidate_idxs[1]] = channel[candidate_idxs[0], candidate_idxs[1]]
        ref_img[ref_img < .3] = 0
        candidate_idxs = np.where(ref_img != 0)
        candidates = list(zip(candidate_idxs[0], candidate_idxs[1]))

        if n < 1 or not misc.isint(n):
            self.sta.status_msg('Skip hot-pixel detection due to wrong settings', self.cfg.params[self.cfg.opt_prnt])
            return channel

        for num, idx in enumerate(candidates):

            j, i = idx

            adj_cands = (candidate_idxs[0] > j-n**2) & (candidate_idxs[0] < j+n**2) & \
                        (candidate_idxs[1] > i-n**2) & (candidate_idxs[1] < i+n**2)

            if np.count_nonzero(adj_cands) < n:

                if n < j < ref_img.shape[0]-n and n < i < ref_img.shape[1]-n:
                    win = channel[j-n:j+n+1, i-n:i+n+1]
                else:
                    # treat candidates being too close to image border
                    alt_n = min(j, abs(ref_img.shape[0]-j), i, abs(ref_img.shape[1]-i))
                    win = channel[j-alt_n:j+alt_n+1, i-alt_n:i+alt_n+1]

                m_val = np.mean(win)
                s_val = np.std(win)

                if channel[j, i] < m_val - s_val * sig_lev or channel[j, i] > m_val + s_val * sig_lev:

                    # replace outlier
                    channel[j, i] = med_img[j, i]

            # check interrupt status
            if self.sta.interrupt:
                return False

        return channel

    @property
    def bay_img(self):
        return self._bay_img



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam import misc
from plenopticam.lfp_extractor.lfp_viewpoints import LfpViewpoints
from plenopticam.lfp_extractor.lfp_exporter import LfpExporter
from plenopticam.lfp_extractor.lfp_contrast import LfpContrast
from plenopticam.misc import GammaConverter

# external
import numpy as np


class LfpShiftAndSum(LfpViewpoints):

    def __init__(self, lfp_img=None, *args, **kwargs):
        super(LfpShiftAndSum, self).__init__(*args, **kwargs)

        # input variables
        self.lfp_img = lfp_img

        # validate refocusing range
        self.validate_range()

        # internal variables
        self._refo_stack = list()

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Compute refocused image stack', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # exclude views that lie outside max. radius for circular blur spots
        self.circular_view_aperture() if self.vp_img_arr is not None else None

        # do refocus computation based on provided method
        if self.vp_img_arr is not None:
            self.refo_from_vp()
        elif self.lfp_img is not None:
            if self.cfg.params[self.cfg.opt_refi]:
                self.refo_from_scratch_upsample()
            else:
                self.refo_from_scratch()

        self._refo_stack = np.asarray(self._refo_stack)

        # tbd
        self.all_in_focus()

        return True

    def refo_from_vp(self):
        """ computational refocusing based on viewpoint shift and integration """

        # print status
        self.sta.progress(0, self.cfg.params[self.cfg.opt_prnt])

        # initialize variables
        self._refo_stack = list()
        patch_len = self.cfg.params[self.cfg.ptc_leng]
        factor = self.cfg.params[self.cfg.ptc_leng] if self.cfg.params[self.cfg.opt_refi] else 1

        # divide intensity to prevent clipping in shift and sum process
        self._vp_img_arr /= patch_len

        # iterate through refocusing parameter a
        ran_refo = self.cfg.params[self.cfg.ran_refo]
        a_list = tuple([factor*a for a in ran_refo]) if self.cfg.params[self.cfg.opt_refi] else ran_refo
        for a in range(*a_list):

            overlap = abs(a) * (patch_len - 1)
            img_slice = np.zeros(np.append(np.array(self._vp_img_arr.shape[2:-1]) * factor + [overlap, overlap], 3))
            for j in range(patch_len):
                for i in range(patch_len):

                    # perform sub-pixel refinement if required
                    vp_img = misc.img_resize(self._vp_img_arr[j, i], factor) if factor > 1 else self._vp_img_arr[j, i]

                    # get viewpoint padding for each border
                    tb = (abs(a) * j, abs(a) * (patch_len - 1 - j))    # top, bottom
                    lr = (abs(a) * i, abs(a) * (patch_len - 1 - i))    # left, right

                    # flip padding for each axis if a is negative
                    pad_width = (tb, lr, (0, 0)) if a >= 0 else (tb[::-1], lr[::-1], (0, 0))

                    # shift viewpoint image and add its values to refocused image slice
                    img_slice = np.add(img_slice, np.pad(vp_img, pad_width, 'edge'))

                    # check interrupt status
                    if self.sta.interrupt:
                        return False

                    # print status
                    percentage = ((j*patch_len+i+1)/patch_len**2+a-a_list[0])/(a_list[-1]-a_list[0])*100
                    self.sta.progress(percentage, self.cfg.params[self.cfg.opt_prnt])

            # crop refocused image for consistent image dimensions
            crop = int(overlap/2)
            final_img = img_slice[crop:-crop, crop:-crop, :] if (a != 0) else img_slice

            # write upscaled version to hard drive
            if self.cfg.params[self.cfg.opt_refi]:
                upscale_img = LfpContrast().auto_hist_align(final_img.copy(), ref_img=final_img, opt=True)
                upscale_img = GammaConverter().srgb_conv(img=upscale_img)
                fname = np.round(a/patch_len, 2)
                LfpExporter(cfg=self.cfg, sta=self.sta).save_refo_slice(fname, upscale_img, string='upscale_')
                del upscale_img

            # spatially downscale image to original resolution (less memory required)
            final_img = misc.img_resize(final_img, 1./factor) if factor > 1 else final_img

            self._refo_stack.append(final_img)

        return True

    def refo_from_scratch(self):    # pragma: no cover
        """ computational refocusing from aligned light field image """

        # print status
        self.sta.progress(0, self.cfg.params[self.cfg.opt_prnt])

        # initialize variables
        self._refo_stack = []
        patch_len = self.cfg.params[self.cfg.ptc_leng]
        if len(self.lfp_img.shape) == 3:
            m, n, P = self.lfp_img.shape
        else:
            m, n, P = self.lfp_img.shape[0], self.lfp_img.shape[1], 1
        J = int(n/patch_len)
        H = int(m/patch_len)
        c = int((patch_len-1)/2)

        # divide intensity to prevent clipping in shift and sum process
        img = self.lfp_img/patch_len

        # start synthesising a new refocused image
        a_range = np.arange(*self.cfg.params[self.cfg.ran_refo])
        for a in a_range:

            # initialize (reset) matrices for intermediate results
            overlap = abs(a) * (patch_len - 1)
            hor_refo = np.zeros([m, J + abs(a) * (patch_len-1), P])
            ver_refo = np.zeros([H + abs(a) * (patch_len-1), J + abs(a) * (patch_len-1), P])

            # horizontal shift and integrate
            for y in range(m):  # jump to next horizontal line
                for j in range(J + overlap):    # each increment is an integration for a refocused pixel
                    for i in range(-c, c+1):    # collects pixels to form refocused pixel
                        u_new = c + i
                        s_new = j + a * (c - i)
                        if a >= 0:
                            s_new -= overlap    # put output image to center
                        x = u_new + s_new * patch_len

                        # if index exceeds image boundaries, leave pixel empty
                        if (x >= 0) & (x < n):
                            fraction_hor = img[y, x, :]
                        else:
                            fraction_hor = np.zeros(P, dtype=self.lfp_img.dtype)

                        # add pixel to refocusing plane (integration)
                        hor_refo[y, j, :] += fraction_hor

            # vertical shift and integrate
            for h in range(H + overlap):
                for g in range(-c, c+1):
                    v_new = c + g
                    t_new = h + a * (c - g)
                    if a >= 0:
                        t_new -= overlap    # put output image to center
                    y = v_new + t_new * patch_len

                    # if index exceeds image boundaries, leave empty
                    if (y >= 0) & (y < m):
                        fraction_ver = hor_refo[y, :, :]
                    else:
                        fraction_ver = np.zeros([J + abs(a) * (patch_len - 1), P], dtype=self.lfp_img.dtype)

                    # add pixel to refocusing plane (integration)
                    ver_refo[h, :, :] += fraction_ver

            # crop refocused image for consistent image dimensions
            crop = int(overlap/2)
            final_img = ver_refo[crop:-crop, crop:-crop, :] if (a != 0) else ver_refo

            # append refocused image to stack
            self._refo_stack.append(final_img)

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status
            self.sta.progress((a-a_range[0]+1)/len(a_range)*100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def refo_from_scratch_upsample(self):   # pragma: no cover

        # print status
        self.sta.progress(0, self.cfg.params[self.cfg.opt_prnt])

        # initialize variables
        self._refo_stack = []
        patch_len = self.cfg.params[self.cfg.ptc_leng]
        m, n, P = self.lfp_img.shape if len(self.lfp_img.shape) == 3 else self.lfp_img.shape[0], self.lfp_img.shape[1], 1
        factor = self.cfg.params[self.cfg.ptc_leng] if self.cfg.params[self.cfg.opt_refi] else 1

        img = self.lfp_img/patch_len    # divide to prevent clipping

        a_range = np.arange(*self.cfg.params[self.cfg.ran_refo])
        for a in a_range:

            # initialization
            overlap = abs(a) * (patch_len - 1)
            hor_refo = np.zeros([m, n + overlap, P])
            ver_refo = np.zeros([m + overlap, n + overlap, P])

            # consider negative shift
            negative_a = overlap if a < 0 else 0

            # horizontal shift and integration
            for y in range(m):
                for x in range(n):

                    # prevent from taking adjacent pixel being beyond image border
                    adj_idx = x + patch_len if x + patch_len < n else x

                    # interpolate values (spatial 1-D) of 2 adjacent pixels for all color channels
                    fraction_vec = np.array([np.ogrid[img[y, x, 0]:img[y, adj_idx, 0]:complex(patch_len)],
                                             np.ogrid[img[y, x, 1]:img[y, adj_idx, 1]:complex(patch_len)],
                                             np.ogrid[img[y, x, 2]:img[y, adj_idx, 2]:complex(patch_len)]]).T

                    # calculate horizontal shift coordinate
                    new_x = x + np.mod(x, patch_len)*(a-1) + negative_a

                    # add fractional values to refocused image row at shifted coordinate
                    hor_refo[y, new_x:new_x + patch_len, :] += fraction_vec

            # vertical shift and integration
            for y in range(m):

                # prevent from taking adjacent pixel being beyond image border
                adj_idx = y + patch_len if y + patch_len < m else y

                frac_vec = np.zeros([patch_len, n + overlap, P])
                for x in range(n + overlap):
                    # interpolate values (spatial 2-D) of 2 adjacent rows for all color channels
                    frac_vec[:, x] = np.array([np.ogrid[hor_refo[y, x, 0]:hor_refo[adj_idx, x, 0]:complex(patch_len)],
                                               np.ogrid[hor_refo[y, x, 1]:hor_refo[adj_idx, x, 1]:complex(patch_len)],
                                               np.ogrid[hor_refo[y, x, 2]:hor_refo[adj_idx, x, 2]:complex(patch_len)]]).T

                # calculate vertical shift coordinate
                new_y = y + np.mod(y, patch_len)*(a-1) + negative_a

                # add fractional values to refocused image at shifted coordinate
                ver_refo[new_y:new_y + patch_len, :, :] += frac_vec

            # crop refocused image for consistent image dimensions
            crop = int(overlap/2)
            final_img = ver_refo[crop:-crop, crop:-crop, :] if (a != 0) else ver_refo

            # write upscaled version to hard drive
            LfpExporter(cfg=self.cfg, sta=self.sta).save_refo_slice(a=a, refo_img=final_img)

            # spatially downscale image to original resolution (for less memory usage)
            final_img = misc.img_resize(final_img, 1./factor) if factor > 1 else final_img

            # append refocused image to stack
            self._refo_stack.append(final_img)

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status
            self.sta.progress((a-a_range[0]+1)/len(a_range)*100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def validate_range(self):

        n = len(self.cfg.params[self.cfg.ran_refo])
        if n == 2:
            one, two = self.cfg.params[self.cfg.ran_refo]
            if one >= two:
                self.cfg.params[self.cfg.ran_refo][1] = one + 1
                self.sta.status_msg('Refocusing range is not positive')
        else:
            raise ValueError("Refocusing range list is supposed to contain only two entries")

    def all_in_focus(self):
        pass

    @property
    def refo_stack(self):
        return self._refo_stack.copy()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.lfp_aligner.lfp_microlenses import LfpMicroLenses
from plenopticam import misc
from plenopticam.misc.type_checks import rint

import numpy as np
from scipy.signal import convolve2d
from color_space_converter import rgb2gry


class LfpDevignetter(LfpMicroLenses):

    def __init__(self, *args, **kwargs):
        super(LfpDevignetter, self).__init__(*args, **kwargs)

        self._wht_img = np.ones(self._lfp_img.shape) if self._wht_img is None else self._wht_img

        # config for decision making whether division by raw image or fit values
        self.noise_lev = kwargs['noise_lev'] if 'noise_lev' in kwargs else None
        self.noise_th = 0.05
        self.patch_mode = False

        # add noise
        self.test = False
        if self.test:
            self._wht_img += np.random.normal(0, .15, self._wht_img.shape)

        # white balance
        if len(self._wht_img.shape) == 3:
            # balance RGB channels in white image
            self._wht_img = rgb2gry(self._wht_img) if self._wht_img.shape[2] == 3 else self._wht_img

        # check for same dimensionality
        self._wht_img = self._wht_img if len(self._wht_img.shape) == len(self._lfp_img.shape) else rgb2gry(self._wht_img)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # normalize white image to upper percentile
        self._wht_img /= np.percentile(self._wht_img, q=99.9)

        # analyse noise in white image
        self.noise_lev = self._estimate_noise_level() if self.noise_lev is None else self.noise_lev

        # print status
        self.sta.status_msg('De-vignetting', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # de-vignetting based on mode and present noise level in white image
        if self.patch_mode and self.noise_lev > self.noise_th:
            # perform fitted white micro image division (to reduce noise propagation)
            self.proc_lens_iter(self.patch_devignetting, prnt=False)
        else:
            # perform ordinary de-vignetting from raw white image (low noise)
            self.wht_img_divide()

        return True

    def wht_img_divide(self):
        """ ordinary devignetting using raw white image division """

        # divide light-field image
        init_arr = np.zeros_like(self._lfp_img)
        self._lfp_img = np.divide(self._lfp_img.copy(), self._wht_img, out=init_arr, where=self._wht_img != 0)

        # status
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def patch_devignetting(self, ly, lx, margin=1):
        """ patch-wise devignetting via white image patch from least-squares regression """

        # find MIC by indices
        mic = self.get_coords_by_idx(ly=ly, lx=lx)

        # slice images
        wht_win = self._extract_win(self._wht_img, mic, margin)
        lfp_win = self._extract_win(self._lfp_img, mic, margin)

        # fit micro image
        if self.noise_lev > self.noise_th:
            _, weight_win = self.fit_patch(wht_win)
        else:
            weight_win = wht_win/wht_win.max()

        # apply vignetting correction
        div_win = np.divide(lfp_win, weight_win, out=np.zeros_like(weight_win), where=weight_win != 0)
        olap = self._cent_pitch + margin
        self._lfp_img[rint(mic[0])-olap:rint(mic[0])+olap+1, rint(mic[1])-olap:rint(mic[1])+olap+1] = div_win

        return True

    def _extract_win(self, img, mic, margin=0):
        win = img[rint(mic[0]) - self._cent_pitch - margin:rint(mic[0]) + self._cent_pitch + margin + 1,
              rint(mic[1]) - self._cent_pitch - margin:rint(mic[1]) + self._cent_pitch + margin + 1]
        return win

    def fit_patch(self, patch):
        """ compute polynomial coefficients and intensity map of 2-D image via least-squares regression """

        x = np.linspace(0, 1, patch.shape[1])
        y = np.linspace(0, 1, patch.shape[0])
        X, Y = np.meshgrid(x, y, copy=False)

        X = X.flatten()
        Y = Y.flatten()
        b = rgb2gry(patch)[..., 0].flatten() if len(patch.shape) == 3 else patch.flatten()

        A = self.compose_vandermonde_2d(X, Y, deg=3)

        # solve for a least squares estimate via pseudo inverse and coefficients in b
        coeffs = np.dot(np.linalg.pinv(A), b)

        # create weighting window
        weight_map = np.dot(A, coeffs).reshape(patch.shape[1], patch.shape[0])
        weight_map /= weight_map.max()

        return coeffs, weight_map

    @staticmethod
    def compose_vandermonde_2d(x, y, deg=2):
        """ compute Vandermonde matrix based on two coordinate dimensions and degree """

        if deg == 1:
            return np.array([np.ones(len(x)), x, y]).T
        elif deg == 2:
            return np.array([np.ones(len(x)), x, y, x * y, x ** 2, y ** 2, x ** 2 * y, x * y ** 2, x ** 2 * y ** 2]).T
        elif deg == 3:
            return np.array([np.ones(len(x)), x, y, x * y, x ** 2, y ** 2, x ** 2 * y, x * y ** 2, x ** 2 * y ** 2,
                             x ** 3, y ** 3, x ** 3 * y, x * y ** 3, x ** 3 * y ** 2, x ** 2 * y ** 3,
                             x ** 3 * y ** 3]).T

    def _estimate_noise_level(self):
        """ estimate white image noise level """

        # print status
        self.sta.status_msg('Estimate white image noise level', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        user_pitch = np.mean(self.cfg.calibs[self.cfg.ptc_mean])
        lp_kernel = misc.create_gauss_kernel(length=int(user_pitch))
        if len(self._wht_img.shape) == 3:
            bw_img = rgb2gry(self._wht_img)[..., 0] if self._wht_img.shape[2] == 3 else self._wht_img[..., 0]
        else:
            bw_img = self._wht_img
        flt_img = convolve2d(bw_img, lp_kernel, 'same')

        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return np.std(bw_img-flt_img)

    @property
    def lfp_img(self):
        return self._lfp_img

    @property
    def wht_img(self):
        return self._wht_img



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import PlenopticamStatus

# external libs
import numpy as np


class NonMaxSuppression(object):

    def __init__(self, img, cfg=None, sta=None):

        # input variables
        self._img = img
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()

        # internal variable
        self._map = np.zeros(self._img.shape, dtype=self._img.dtype)

    @property
    def map(self):
        return self._map

    @property
    def idx(self):
        return np.array(self._map.nonzero())

    def main(self):

        # ckeck interrupt status
        if self.sta.interrupt:
            return False

        # find local maxima
        self._non_max_suppression()

        # suppress negative local maxima
        self._map[self._map < 0] = 0

    def _non_max_suppression(self):
        """ adaption of non-maximum suppression by Tuan Q. Pham """

        # print status
        self.sta.status_msg('Select maxima', self.cfg.params[self.cfg.opt_prnt])

        h, w = self._img.shape
        skip = np.zeros([h, 2], dtype=self._img.dtype)
        cur = 0
        next = 1 # scanline masks
        for c in range(1, w-1):
            r = 1
            while r < h-1:
                # skip current pixel
                if skip[r, cur]:
                    r += 1
                    continue
                # compare to pixel on the left
                if self._img[r, c] <= self._img[r+1, c]:
                    r += 1
                    # rising
                    while r < h-1 and self._img[r, c] <= self._img[r+1, c]:
                        r += 1
                    if r == h-1:
                        # reach scanline's local maximum
                        break
                else:
                    # compare to pixel on the right
                    if self._img[r, c] <= self._img[r-1, c]:
                        r += 1
                        continue
                skip[r+1, cur] = 1

                # compare to 3 future then 3 past neighbors
                if self._img[r, c] <= self._img[r-1, c+1]:
                    r += 1
                    continue
                # skip future neighbors only
                skip[r-1, next] = 1
                if self._img[r, c] <= self._img[r, c+1]:
                    r += 1
                    continue
                skip[r, next] = 1
                if self._img[r, c] <= self._img[r+1, c+1]:
                    r += 1
                    continue
                skip[r+1, next] = 1
                if self._img[r, c] <= self._img[r-1, c-1]:
                    r += 1
                    continue
                if self._img[r, c] <= self._img[r, c-1]:
                    r += 1
                    continue
                if self._img[r, c] <= self._img[r+1, c-1]:
                    r += 1
                    continue

                self._map[r, c] = self._img[r, c]
                # a new local maximum is found
                r += 1

            tmp = cur
            cur = next
            # swap mask indices
            next = tmp
            # reset next scanline mask
            skip[:, next] = 0

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print status
            self.sta.progress(c/(w-2)*100, self.cfg.params[self.cfg.opt_prnt])

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import unittest
from os.path import exists, join, basename, splitext

from plenopticam.lfp_calibrator import LfpCalibrator
from plenopticam.lfp_aligner import LfpAligner
from plenopticam.lfp_extractor import LfpExtractor
from plenopticam.lfp_refocuser import LfpRefocuser
from plenopticam.cfg import PlenopticamConfig, constants
from plenopticam.misc import PlenopticamStatus, mkdir_p, load_img_file
from plenopticam.misc.data_downloader import DataDownloader


class PlenoptiCamTesterCustom(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(PlenoptiCamTesterCustom, self).__init__(*args, **kwargs)

    def setUp(self):

        # set config for unit test purposes
        self.sta = PlenopticamStatus()
        self.cfg = PlenopticamConfig()
        self.cfg.reset_values()
        self.cfg.params[self.cfg.opt_dbug] = False
        self.cfg.params[self.cfg.opt_prnt] = False    # prevent Travis CI to terminate after reaching 4MB logfile size
        self.cfg.params[self.cfg.opt_vign] = False
        self.cfg.params[self.cfg.opt_sat_] = True
        self.cfg.params[self.cfg.cal_meth] = constants.CALI_METH[3]
        self.cfg.params[self.cfg.opt_cali] = True

        # retrieve OpEx data from Hahne et al.
        self.loader = DataDownloader(cfg=self.cfg, sta=self.sta)
        self.fp = join(self.loader.root_path, 'examples', 'data')
        archive_fn = join(self.fp, basename(self.loader.opex_url))
        self.loader.download_data(self.loader.opex_url, fp=self.fp) if not exists(archive_fn) else None
        self.loader.extract_archive(archive_fn, fname_list=self.loader.opex_fnames_wht + self.loader.opex_fnames_lfp)

    def test_custom_cal(self):

        for fn_lfp, fn_wht in zip(self.loader.opex_fnames_lfp, self.loader.opex_fnames_wht):

            # generate console output to prevent abort in Travis CI
            print(fn_wht)

            # update file paths and calibration data in config
            self.cfg.params[self.cfg.lfp_path] = join(self.fp, fn_lfp)
            self.cfg.params[self.cfg.cal_path] = join(self.fp, fn_wht)

            # create folder (if it doesn't already exist)
            mkdir_p(splitext(self.cfg.params[self.cfg.lfp_path])[0])

            # test light field calibration
            wht_img = load_img_file(self.cfg.params[self.cfg.cal_path])
            cal_obj = LfpCalibrator(wht_img=wht_img, cfg=self.cfg, sta=self.sta)
            ret_val = cal_obj.main()
            del cal_obj

            # assertion
            self.assertEqual(True, ret_val)

    def test_custom_lfp(self):

        for fn_lfp, fn_wht in zip(self.loader.opex_fnames_lfp, self.loader.opex_fnames_wht):

            # generate console output to prevent abort in Travis CI
            print(fn_lfp)

            # update file paths and calibration data in config
            self.cfg.params[self.cfg.lfp_path] = join(self.fp, fn_lfp)
            self.cfg.params[self.cfg.cal_path] = join(self.fp, fn_wht)
            self.cfg.params[self.cfg.cal_meta] = splitext(self.cfg.params[self.cfg.cal_path])[0]+'.json'
            self.cfg.load_cal_data()

            # create folder (if it doesn't already exist)
            mkdir_p(splitext(self.cfg.params[self.cfg.lfp_path])[0])

            # test light field alignment
            lfp_img = load_img_file(self.cfg.params[self.cfg.lfp_path])
            lfp_obj = LfpAligner(lfp_img=lfp_img, cfg=self.cfg, sta=self.sta)
            ret_val = lfp_obj.main()
            lfp_img = lfp_obj.lfp_img
            del lfp_obj

            # assertion
            self.assertEqual(True, ret_val)

            # test light field extraction
            lfp_obj = LfpExtractor(lfp_img_align=lfp_img, cfg=self.cfg, sta=self.sta)
            ret_val = lfp_obj.main()
            vp_img_arr = lfp_obj.vp_img_arr
            del lfp_obj

            # assertion
            self.assertEqual(True, ret_val)

            lfp_obj = LfpRefocuser(vp_img_arr=vp_img_arr, cfg=self.cfg, sta=self.sta)
            ret_val = lfp_obj.main()
            del lfp_obj

            # assertion
            self.assertEqual(True, ret_val)

    def test_all(self):

        self.test_custom_cal()
        self.test_custom_lfp()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import unittest
import numpy as np
from depthy.misc.pfm_handler import load_pfm

from plenopticam.lfp_extractor.lfp_depth import LfpDepth
from plenopticam.misc import load_img_file, PlenopticamError


class PlenopticamTesterPlt(unittest.TestCase):

    def setUp(self):

        self.depth_map = np.ones([6, 6])
        self.lfp_depth_obj = LfpDepth(depth_map=self.depth_map)

        self.plot_opt = False

    def test_real_data(self):

        self.depth_map, _ = load_pfm('../examples/data/gradient_rose_close/depth.pfm')
        self.rgb_img = load_img_file('../examples/data/gradient_rose_close/thumbnail.png')
        self.lfp_depth_obj = LfpDepth(depth_map=self.depth_map)

        # test invalid downscale parameters
        for i in range(-1, 5):
            print(i)
            try:
                self.lfp_depth_obj.plot_point_cloud(rgb_img=self.rgb_img, down_scale=i)
            except PlenopticamError as e:
                self.assertTrue(e, PlenopticamError)

        # test Axes3D argument
        import matplotlib.pyplot as plt
        fig, ax = plt.figure(), plt.axes(projection='3d')
        self.lfp_depth_obj.plot_point_cloud(down_scale=4, view_angles=(50, 70), ax=ax)

    def test_point_cloud(self):

        # test case where rgb image missing
        self.lfp_depth_obj.plot_point_cloud(rgb_img=None, down_scale=1)

        # test case for image dimension mismatch
        self.lfp_depth_obj.plot_point_cloud(rgb_img=np.zeros([3, 3, 3]), down_scale=1)

        # test valid case
        self.lfp_depth_obj.plot_point_cloud(rgb_img=np.zeros_like(self.depth_map), down_scale=2)

    def test_all(self):

        self.test_point_cloud()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# external libs
import numpy as np

from plenopticam.lfp_extractor import LfpViewpoints
from plenopticam.lfp_aligner.cfa_processor import CfaProcessor

try:
    from scipy.signal import medfilt
    from scipy.ndimage import median_filter
except ImportError:
    raise ImportError('Please install scipy package.')

from color_matcher import ColorMatcher


class LfpColorEqualizer(LfpViewpoints):

    def __init__(self, *args, **kwargs):
        super(LfpColorEqualizer, self).__init__(*args, **kwargs)

        self._ref_img = kwargs['ref_img'] if 'ref_img' in kwargs else self.central_view
        self.prop_type = kwargs['prop_type'] if 'prop_type' in kwargs else 'central'
        self._method = 'hm-mkl-hm'

        # skip color propagation if option not set
        self.prop_type = self.prop_type if hasattr(self, 'cfg') and self.cfg.params[self.cfg.opt_colo] else None

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # apply color correction
        if self.vp_img_arr is not None:
            self.apply_ccm()
            self._ref_img = self.central_view

        # equalize light field colors
        if self.prop_type == 'central':
            self.proc_vp_arr(fun=self.color_eq_img, ref=self._ref_img, method=self._method, msg='Color equalization')
        elif self.prop_type == 'axial':
            self.proc_ax_propagate_2d(fun=self.color_eq_img, method=self._method, msg='Color equalization')

        # zero-out marginal sub-apertures (e.g. suffering from cross-talk)
        self._exclude_crosstalk_views()

    @staticmethod
    def color_eq_img(src, ref, method=None):

        # instantiate color matcher
        match = ColorMatcher(src, ref, method=method).main()

        return match

    def apply_ccm(self):

        # color matrix correction
        if 'ccm' in self.cfg.lfpimg.keys():

            # ccm mat selection
            if 'ccm_wht' in self.cfg.lfpimg:
                ccm_arr = self.cfg.lfpimg['ccm_wht']
            elif 'ccm' in self.cfg.lfpimg:
                # hard-coded and generic CCM for Lytro
                ccm_arr = np.array([2.4827811717987061, -1.1018080711364746, -0.38097298145294189,
                                    -0.36761483550071716, 1.6667767763137817, -0.29916191101074219,
                                    -0.18722048401832581, -0.73317205905914307, 1.9203925132751465])
                # ccm_arr = self.cfg.lfpimg['ccm']
            else:
                ccm_arr = np.diag(np.ones(3))

            # normalize
            self.vp_img_arr /= self.vp_img_arr.max()

            if 'exp' in self.cfg.lfpimg:
                sat_lev = 2 ** (-self.cfg.lfpimg['exp'])
            else:
                sat_lev = 1
            self.vp_img_arr *= sat_lev

            # transpose and flip ccm_mat for RGB order
            ccm_mat = np.array(ccm_arr).reshape(3, 3).T
            self._vp_img_arr = CfaProcessor().correct_color(self._vp_img_arr.copy(), ccm_mat=ccm_mat)

            # remove potential NaNs
            self._vp_img_arr[self._vp_img_arr < 0] = 0
            self._vp_img_arr[self._vp_img_arr > sat_lev] = sat_lev
            self._vp_img_arr /= self._vp_img_arr.max()

        return True

    def _exclude_crosstalk_views(self):
        """ function wrapper to exclude Lytro Illum views that suffer from cross-talk """

        self.circular_view_aperture(offset=2, ellipse=True)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


# local imports
from plenopticam import misc
from plenopticam.misc.errors import LfpTypeError, PlenopticamError
from plenopticam.lfp_reader.lfp_decoder import LfpDecoder
from plenopticam.lfp_reader.constants import SUPP_FILE_EXT
from plenopticam.misc.gamma_converter import GammaConverter

import os


class LfpReader(object):

    def __init__(self, cfg=None, sta=None, lfp_path=None):

        # input and output variables
        self.cfg = cfg
        self.sta = sta if sta is not None else misc.PlenopticamStatus()

        # internal variables
        self._lfp_path = lfp_path if lfp_path is not None else cfg.params[cfg.lfp_path]

        # output variables
        self._bay_img = None
        self._lfp_img = None
        self._json_dict = None

        # filename and file path from previously decoded data
        self.dp = os.path.splitext(self._lfp_path)[0]
        self.fn = os.path.basename(self.dp) + '.tiff'
        self.fp = os.path.join(self.dp, self.fn)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        if self._lfp_path.lower().endswith(SUPP_FILE_EXT):

            try:
                self.decode_lytro_file()
            except FileNotFoundError:
                # print status
                self.sta.status_msg('{0} not found'.format(os.path.basename(self._lfp_path)), self.cfg.params[self.cfg.opt_prnt])
                self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])
                self.sta.error = True
            except Exception as e:
                # unrecognized LFP file type
                if not self._json_dict:
                    raise LfpTypeError(e, cfg=self.cfg)
                else:
                    raise PlenopticamError(e, cfg=self.cfg, sta=self.sta)
        else:
            try:
                # read and decode generic image file type
                self._lfp_img = misc.load_img_file(self._lfp_path)
                # inverse sRGB conversion
                self._lfp_img = GammaConverter().srgb_conv(self._lfp_img, inverse=True)
            except TypeError:
                self.sta.status_msg('File type not recognized')
                self.sta.error = True
                return False

            try:
                # try to load json file (if present)
                json_dict = self.cfg.load_json(self._lfp_path)
                self.cfg.lfpimg = LfpDecoder.filter_lfp_json(json_dict, self.cfg.lfpimg)
            except (FileNotFoundError, AttributeError):
                pass

        # write json file
        self.cfg.save_params()

        return True

    def decode_lytro_file(self):

        # Lytro type decoding
        with open(self._lfp_path, mode='rb') as file:

            # LFC and raw type decoding
            obj = LfpDecoder(file, self.cfg, self.sta, lfp_path=self._lfp_path)
            obj.main()
            self._lfp_img = obj.bay_img
            self._json_dict = obj.json_dict
            del obj

            # save bayer image as file (skip if it exists to save time)
            if not os.path.exists(self.fp) and not self.sta.interrupt:
                self.sta.status_msg(msg='Save raw image', opt=self.cfg.params[self.cfg.opt_prnt])
                self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])
                misc.save_img_file(misc.Normalizer(self._lfp_img).uint16_norm(), self.fp, file_type='tiff')
                self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    @property
    def lfp_img(self):
        return self._lfp_img



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import os

try:
    from PIL import Image
except ImportError:
    raise ImportError('Please install pillow.')

from plenopticam.misc.normalizer import Normalizer
from plenopticam.misc import suppress_user_warning

PORTISHEAD = b"x\x9cm\x8f\xe1\n\xc0 \x08\x84\xdf\xffEu\x8c\x84`kBM\x9d\x95\xc4`\xbb?\xde\xa7R\x9e\x99K\xa55Q\x0b)" + \
             b"\x13\x02 \xf1\xecH\x86P\x96>]\xe8\r\xdf\xe0nRJ[\xaflJ^P\xb8\xdc\xc9\r\xa9\xe0\xe0\x1d\xcek\x98\x06" + \
             b"\xc1|t\xd7\x82E\n\x0e^\xfb0\x07\xf1^0i\xfc\x87\x93\xf9{\xcf\xfb^\xfd\xcb3\xf2\xd6\x1ay\x1f\xc8\x93\xf0u"


def try_tiff_import(type):

    try:
        import imageio
    except:
        imageio, type = None, 'png'

    return imageio, type


def save_img_file(img, file_path=None, file_type=None, gamma=None, tag=None):

    # do gamma correction
    img = img**gamma if gamma is not None else img

    file_path = os.getcwd() if file_path is None else file_path

    try:
        img = place_dnp(img) if not tag else img
    except ValueError:
        pass

    # amend write privileges of (potentially existing) config file
    if os.path.exists(file_path):
        st = os.stat(file_path)
        os.chmod(file_path, st.st_mode | 0o111)

    ext = os.path.splitext(file_path)[-1][1:]
    if not file_type:
        file_type = ext if ext == 'png' or ext == 'tiff' else 'tiff' if img.dtype == 'uint16' else 'png'

    # try imageio import or use png instead if import fails
    imageio, file_type = try_tiff_import(file_type)

    # compose new file path string if extension type changed
    file_path = os.path.splitext(file_path)[-2] if file_path.endswith(('.tiff', '.png', '.bmp')) else file_path
    file_type = 'png' if file_type is None else file_type
    file_path += '.' + file_type

    if file_type.__contains__('tif'):
        suppress_user_warning(True, category=UserWarning)
        imageio.imwrite(uri=file_path, im=Normalizer(img).uint16_norm())
        suppress_user_warning(False, category=UserWarning)

    elif file_type == 'png' or file_type == 'bmp':
        try:
            #Image.fromarray(Normalizer(img).uint8_norm()).save(file_path, file_type, optimize=True)
            imageio.imwrite(uri=file_path, im=Normalizer(img).uint8_norm())
        except PermissionError as e:
            raise Exception(e)

    return True


def load_img_file(file_path):

    file_type = file_path.split('.')[-1]

    # try libtiff import or use png instead if import fails
    imageio, file_type = try_tiff_import(file_type)

    if file_type.__contains__('tif'):
        suppress_user_warning(True, category=UserWarning)
        img = imageio.imread(uri=file_path)
        suppress_user_warning(False, category=UserWarning)

    elif any(file_type in ext for ext in ('bmp', 'png', 'jpeg', 'jpg')):
        try:
            img = Image.open(file_path)
            #imageio.imread(uri=file_path, format=file_type)
        except OSError:
            # support load of truncated images
            from PIL import ImageFile
            ImageFile.LOAD_TRUNCATED_IMAGES = True
            img = Image.open(file_path)
        except AttributeError:
            raise TypeError('Filetype %s not recognized' % file_type)

    else:
        raise TypeError('Filetype %s not recognized' % file_type)

    # normalize (convert to numpy array)
    img = Normalizer(np.asarray(img)).type_norm()

    return img


def save_gif(img_set, duration=.1, fp='', fn='default'):

    fn = '%s.gif' % fn

    try:
        import imageio
        imageio.mimwrite(os.path.join(fp, fn), [place_dnp(img) for img in img_set], duration=duration, palettesize=2**8)
    except ImportError:
        # only use pillow for gif animation if necessary as it yields poor image quality
        pil_arr = [Image.fromarray(place_dnp(img)) for img in img_set]
        pil_arr[0].save(os.path.join(fp, fn), save_all=True, append_images=pil_arr[1:], duration=duration, loop=0)
    except PermissionError:
        pass

    return True


def place_dnp(dus):

    import zlib

    dtype = dus.dtype
    dus = dus.astype('float64')

    n = 16
    dnp = np.asarray(np.frombuffer(zlib.decompress(PORTISHEAD), 'uint8').reshape(n, n))/(2**8-1.)

    s, t = dus.shape[:2]
    y, x = np.array([s-n, s+(t-s)//2-n])-n//2
    a = dus.shape[2] if len(dus.shape) == 3 else 1
    dus[y:y+n, x:x+n, ...] *= np.repeat(dnp[..., np.newaxis], a, axis=2) if a > 1 else dnp

    return dus.astype(dtype)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import sys
import unittest
import subprocess
import time


class ExecutableTester(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(ExecutableTester, self).__init__(*args, **kwargs)

    def setUp(self):
        pass

    def run_app(self):

        if sys.platform == 'linux':
            cmd = 'dist/plenopticam'
        elif sys.platform == 'darwin':
            cmd = 'open dist/plenopticam.app'
        elif sys.platform == 'win32':
            cmd = 'dist\plenopticam.exe'
        else:
            cmd = 'dist/plenopticam'

        process = subprocess.Popen("exec " + cmd, stdout=subprocess.PIPE, shell=True)   # , preexec_fn=os.setsid
        time.sleep(20)
        process.kill()

        result, error = process.communicate()
        self.assertEqual(b'', result)
        self.assertEqual(None, error)

    def test_all(self):

        self.run_app()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

import os

from plenopticam.gui.widget_path import PathWidget
from plenopticam.gui.constants import PX, PY, GEN_IMG_EXTS, LYT_LFP_EXTS, LYT_CAL_EXTS, ALL_EXTS


class FileWidget(tk.Frame):

    def __init__(self, parent):

        # inheritance
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self.cfg = parent.cfg

        # supported file extensions
        gen_img_exts = ['*.'+ext for ext in GEN_IMG_EXTS]
        lyt_lfp_exts = ['*.'+ext for ext in LYT_LFP_EXTS]
        lyt_cal_exts = ['*.'+ext for ext in LYT_CAL_EXTS]
        merge_lfp_exts = lyt_lfp_exts+gen_img_exts
        merge_cal_exts = lyt_cal_exts+gen_img_exts
        LFP_EXTS = [('Supported files', ' '.join(merge_lfp_exts+[x.upper() for x in merge_lfp_exts])),
                    ('Lytro files', ' '.join(lyt_lfp_exts+[x.upper() for x in lyt_lfp_exts])),
                    ('Generic image files', ' '.join(gen_img_exts+[x.upper() for x in gen_img_exts])),
                    ('All files', ' '.join(ALL_EXTS+[x.upper() for x in ALL_EXTS]))]
        CAL_EXTS = [('Supported files', ' '.join(merge_cal_exts+[x.upper() for x in merge_cal_exts])),
                    ('Lytro files', ' '.join(LYT_CAL_EXTS+[x.upper() for x in lyt_cal_exts])),
                    ('Generic image files', ' '.join(gen_img_exts+[x.upper() for x in gen_img_exts])),
                    ('All files', ' '.join(ALL_EXTS+[x.upper() for x in ALL_EXTS]))]

        # instantiate light field path widget
        tk.Label(self, text='Light field image').grid(row=0, column=0, sticky='W')
        self.lfp_wid = PathWidget(self, path=self.cfg.params[self.cfg.lfp_path], path_type=False, file_exts=LFP_EXTS)
        self.lfp_wid.grid(row=0, column=1, padx=PX, pady=PY)
        self.lfp_wid.bind_to(self.set_lfp_path)     # observe change in path variable

        # instantiate calibration path widget
        tk.Label(self, text='Calibration source').grid(row=1, column=0, sticky='W')
        self.cal_wid = PathWidget(self, path=self.cfg.params[self.cfg.cal_path], path_type=False, file_exts=CAL_EXTS)
        self.cal_wid.grid(row=1, column=1, padx=PX, pady=PY)
        self.cal_wid.bind_to(self.set_cal_path)     # observe change in path variable

        # radio button to enable change from path to file type
        self.cal_wid.path_type = os.path.isdir(self.cfg.params[self.cfg.cal_path])
        self.chk_var = tk.BooleanVar(value=bool(self.cal_wid.path_type))
        self.chk_btn = tk.Checkbutton(self, text='Folder', variable=self.chk_var, command=self.btn_update)
        self.chk_btn.grid(row=1, column=2, sticky='W')

        # list of button and entry widgets (collected to disable/enable widgets)
        self.btn_list = [self.lfp_wid.btn, self.cal_wid.btn, self.chk_btn, self.lfp_wid.ent, self.cal_wid.ent]

    def btn_update(self):
        # toggle path type in PathWidget for calibration
        self.cal_wid.path_type = not self.cal_wid.path_type

    def set_lfp_path(self, val):

        self.cfg.params[self.cfg.lfp_path] = val
        self.cfg.save_params()

    def set_cal_path(self, val):

        self.cfg.params[self.cfg.cal_path] = val
        self.cfg.save_params()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import PlenopticamStatus
from plenopticam.misc.circle_drawer import bresenham_circle

import numpy as np


class LfpViewpoints(object):

    def __init__(self, *args, **kwargs):

        self._vp_img_arr = kwargs['vp_img_arr'] if 'vp_img_arr' in kwargs else None
        self._vp_img_arr = self.vp_img_arr.astype('float64') if self.vp_img_arr is not None else None
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else PlenopticamConfig()
        self.sta = kwargs['sta'] if 'sta' in kwargs else PlenopticamStatus()
        self._size_pitch = self.cfg.params[self.cfg.ptc_leng]
        self._cent_pitch = self._size_pitch // 2

        try:
            self._DIMS = self._vp_img_arr.shape if len(self._vp_img_arr.shape) == 3 else self._vp_img_arr.shape + (1,)
        except (TypeError, AttributeError):
            pass
        except IndexError:
            self.sta.status_msg('Incompatible image dimensions: Please either use KxLx3 or KxLx1 array dimensions')
            self.sta.error = True

    @property
    def vp_img_arr(self):
        return self._vp_img_arr

    @vp_img_arr.setter
    def vp_img_arr(self, vp_img_arr):
        self._vp_img_arr = vp_img_arr

    @property
    def central_view(self):
        return self._vp_img_arr[self._cent_pitch, self._cent_pitch, ...].copy() if self._vp_img_arr is not None else None

    @staticmethod
    def remove_proc_keys(kwargs, data_type=None):

        data_type = dict if not data_type else data_type
        keys_to_remove = ('cfg', 'sta', 'msg', 'iter_num', 'iter_tot')

        if data_type == dict:
            output = dict((key, kwargs[key]) for key in kwargs if key not in keys_to_remove)
        elif data_type == list:
            output = list(kwargs[key] for key in kwargs.keys() if key not in keys_to_remove)
        else:
            output = None

        return output

    def proc_vp_arr(self, fun, **kwargs):
        """ process viewpoint images based on provided function handle and argument data """

        # percentage indices for tasks having sub-processes
        iter_num = kwargs['iter_num'] if 'iter_num' in kwargs else 0
        iter_tot = kwargs['iter_tot'] if 'iter_tot' in kwargs else 1

        # status message handling
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])
        if iter_num == 0:
            msg = kwargs['msg'] if 'msg' in kwargs else 'Viewpoint process'
            self.sta.status_msg(msg, self.cfg.params[self.cfg.opt_prnt])

        args = self.remove_proc_keys(kwargs, data_type=list)

        # light-field shape handling
        if len(self.vp_img_arr.shape) != 5:
            raise NotImplementedError
        new_shape = fun(self._vp_img_arr[0, 0, ...].copy(), *args).shape
        new_array = np.zeros(self._vp_img_arr.shape[:2] + new_shape)

        for j in range(self._vp_img_arr.shape[0]):
            for i in range(self._vp_img_arr.shape[1]):

                # check interrupt status
                if self.sta.interrupt:
                    return False

                res = fun(self._vp_img_arr[j, i, ...], *args)

                if res.shape == self._vp_img_arr.shape:
                    self._vp_img_arr[j, i, ...] = res
                else:
                    new_array[j, i, ...] = res

                # progress update
                percent = (j*self._vp_img_arr.shape[1]+i+1)/np.dot(*self._vp_img_arr.shape[:2])
                percent = percent / iter_tot + iter_num / iter_tot
                self.sta.progress(percent*100, self.cfg.params[self.cfg.opt_prnt])

        if new_array.sum() != 0:
            self._vp_img_arr = new_array

        return True

    @staticmethod
    def get_move_coords(arr_dims: (int, int) = (None, None), pattern: str = None, r: int = None) -> list:
        """ compute view coordinates that are used for loop iterations """

        # parameter initialization
        pattern = 'circle' if pattern is None else pattern
        r = r if r is not None else min(arr_dims)//2
        mask = [[0] * arr_dims[1] for _ in range(arr_dims[0])]

        if pattern == 'square':
            mask[0, :] = 1
            mask[:, 0] = 1
            mask[-1, :] = 1
            mask[:, -1] = 1
        if pattern == 'circle':
            mask = bresenham_circle(arr_dims, r=r)

        # extract coordinates from mask
        coords_table = [(y, x) for y in range(len(mask)) for x in range(len(mask)) if mask[y][x]]

        # sort coordinates in angular order
        coords_table.sort(key=lambda coords: np.arctan2(coords[0] - arr_dims[0]//2, coords[1] - arr_dims[1]//2))

        return coords_table

    def reorder_vp_arr(self, pattern=None, lf_radius=None):

        # parameter initialization
        pattern = 'circle' if pattern is None else pattern
        move_coords = self.get_move_coords(arr_dims=self.vp_img_arr.shape[:2], pattern=pattern, r=lf_radius)

        vp_img_set = []
        for coords in move_coords:
            vp_img_set.append(self.vp_img_arr[coords[0], coords[1], ...])

        return vp_img_set

    def proc_ax_propagate_1d(self, fun, idx=None, axis=None, **kwargs):
        """ apply provided function along axis direction """

        # status message handling
        if 'msg' in kwargs:
            self.sta.status_msg(kwargs['msg'], self.cfg.params[self.cfg.opt_prnt])

        axis = 0 if axis is None else axis
        j = 0 if idx is None else idx
        m, n = (0, 1) if axis == 0 else (1, 0)
        p, q = (1, -1) if axis == 0 else (-1, 1)

        for i in range(self._cent_pitch):

            # swap axes indices
            j, i = (i, j) if axis == 1 else (j, i)

            ref_pos = self.vp_img_arr[self._cent_pitch+j, self._cent_pitch+i, ...]
            ref_neg = self.vp_img_arr[self._cent_pitch+j*p, self._cent_pitch+i*q, ...]

            self._vp_img_arr[self._cent_pitch+j+m, self._cent_pitch+i+n, ...] = \
                fun(self.vp_img_arr[self._cent_pitch+j+m, self._cent_pitch+i+n, ...], ref_pos, **kwargs)
            self._vp_img_arr[self._cent_pitch+(j+m)*p, self._cent_pitch+(i+n)*q, ...] = \
                fun(self.vp_img_arr[self._cent_pitch+(j+m)*p, self._cent_pitch+(i+n)*q, ...], ref_neg, **kwargs)

            # swap axes indices
            j, i = (i, j) if axis == 1 else (j, i)

            # check interrupt status
            if self.sta.interrupt:
                return False

        return True

    def proc_ax_propagate_2d(self, fun, **kwargs):
        """ apply provided function along axes """

        # percentage indices for tasks having sub-processes
        iter_num = kwargs['iter_num'] if 'iter_num' in kwargs else 0
        iter_tot = kwargs['iter_tot'] if 'iter_tot' in kwargs else 1

        # status message handling
        if iter_num == 0:
            msg = kwargs['msg'] if 'msg' in kwargs else 'Viewpoint process'
            self.sta.status_msg(msg, self.cfg.params[self.cfg.opt_prnt])

        kwargs = self.remove_proc_keys(kwargs, data_type=dict)

        self.proc_ax_propagate_1d(fun, idx=0, axis=0, **kwargs)

        for j in range(-self._cent_pitch, self._cent_pitch+1):

            # apply histogram matching along entire column
            self.proc_ax_propagate_1d(fun, idx=j, axis=1, **kwargs)

            # progress update
            percent = (j+self._cent_pitch+1) / self._vp_img_arr.shape[0]
            percent = percent / iter_tot + iter_num / iter_tot
            self.sta.progress(percent*100, self.cfg.params[self.cfg.opt_prnt])

            # check interrupt status
            if self.sta.interrupt:
                return False

        return True

    @property
    def views_stacked_img(self):
        """ 4D to 2D flattening of all sub-aperture images for a single image representation """
        return np.moveaxis(np.concatenate(np.moveaxis(np.concatenate(np.moveaxis(self.vp_img_arr, 1, 2)), 0, 2)), 0, 1)

    def circular_view_aperture(self, offset=None, ellipse=None):

        # initialize variables
        offset = offset if offset is not None else 0
        ratio = self.vp_img_arr.shape[3]/self.vp_img_arr.shape[2] if ellipse else 1
        r = self._size_pitch // 2
        mask = np.zeros([2*r+1, 2*r+1])

        # determine mask for affected views
        for x in range(-r, r + 1):
            for y in range(-r, r + 1):
                if int(np.round(np.sqrt(x ** 2 + y ** 2 * ratio))) > r + offset:
                    mask[r + y][r + x] = 1

        # extract coordinates from mask
        coords_table = [(y, x) for y in range(len(mask)) for x in range(len(mask)) if mask[y][x]]

        # zero-out selected views
        for coords in coords_table:
            self.vp_img_arr[coords[0], coords[1], ...] = np.zeros(self.vp_img_arr.shape[2:])

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from plenopticam import __version__
from plenopticam.gui.constants import PX, PY, BTN_W


# make object for config window
class AbtWidget(object):

    def __init__(self):

        # open about window
        self.abt_widget = tk.Toplevel(padx=PX, pady=PY)

        # window title
        self.abt_widget.wm_title("About")

        # name and version
        line = tk.Label(master=self.abt_widget, text="Plenopticam "+__version__)
        line.grid(row=0, padx=PX, pady=PY, sticky='NWSE')

        # website
        link = tk.Label(master=self.abt_widget, text=r"http://www.plenoptic.info")
        link.grid(row=1, padx=PX, pady=PY, sticky='NWSE')

        # license
        license = tk.Label(master=self.abt_widget, text=__license__)
        license.grid(row=2, padx=PX, sticky='NWSE')

        btn = tk.Button(self.abt_widget, text="Close", width=BTN_W, command=self.close)
        btn.grid(row=3, padx=PX, pady=PY)

    def close(self):
        """ close about window """

        self.abt_widget.destroy()

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


def find_centroid(centroids, ref_point, pitch, axis, pattern='hex', odd=True, e=3, inv_dir=False):

    # set indices for horizontal or vertical search respectively
    i, j = [1, 0] if axis == 1 else [0, 1]

    # consider hexagonal shift alternation which is expected along vertical axis=0
    h, g = [1, 1] if pattern == 'rec' or axis == 1 else [0, e/2] if odd else [e/2, 0]

    # set search condition
    if not inv_dir:
        # forward condition for left and down direction
        cond = (centroids[:, i] > ref_point[i] + pitch[i]/2) & (centroids[:, i] < ref_point[i]+e*pitch[i]/2) & \
               (centroids[:, j] > ref_point[j]-h*pitch[j]/2) & (centroids[:, j] < ref_point[j]+g*pitch[j]/2)
    else:
        # backward condition for right and up direction
        cond = (centroids[:, i] < ref_point[i] - pitch[i]/2) & (centroids[:, i] > ref_point[i]-e*pitch[i]/2) & \
               (centroids[:, j] > ref_point[j]-h*pitch[j]/2) & (centroids[:, j] < ref_point[j]+g*pitch[j]/2)

    # find centroid given the condition
    found_centroid = centroids[cond].ravel()

    return found_centroid



#!/usr/bin/env python
import os

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# external libs
import numpy as np
from color_space_converter import rgb2gry

# local imports
from plenopticam.misc import Normalizer, save_img_file, PlenopticamStatus
from plenopticam.cfg import PlenopticamConfig


class CentroidDrawer(object):

    def __init__(self, img, centroids, cfg=None, sta=None):

        # input variables
        self._img = rgb2gry(img.copy())[..., 0] if len(img.shape) == 3 else img.copy()
        self._img = Normalizer(self._img).uint8_norm()
        self._centroids = np.asarray(centroids)
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()

    def write_centroids_img(self, fn='centroids_img.png'):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # status message handling
        self.sta.status_msg(msg='Save centroids image', opt=self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # draw centroids in binary image and save image file for debug purposes
        plot_img = self.draw_centroids_img()
        save_img_file(plot_img, os.path.join(os.path.splitext(self.cfg.params[self.cfg.lfp_path])[0], fn))

        # status message handling
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def draw_centroids_img(self):

        # draw centroids in single channel image
        centroid_mask = np.zeros(self._img.shape[:2], dtype=np.uint8)
        idx = np.round(self._centroids[:, 0]).astype(int), np.round(self._centroids[:, 1]).astype(int)
        centroid_mask[idx] = np.iinfo(centroid_mask.dtype).max

        # merge image and centroids to form RGB image
        diff_mask = self.non_wrap_subtract(self._img, centroid_mask)
        plot_img = np.zeros([self._img.shape[0], self._img.shape[1], 3], dtype=np.uint8)
        plot_img[..., :2] = np.concatenate((diff_mask[..., np.newaxis], diff_mask[..., np.newaxis]), axis=2)
        plot_img[..., 2] = centroid_mask            # move centroid_mask to blue channel

        return plot_img

    @staticmethod
    def non_wrap_subtract(a, b):
        """ prevent wrap-around for np.ndarray """

        x = (a - b)
        x[b > a] = 0

        return x



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except ImportError:
    import Tkinter as tk
    import Tkinter.ttk as ttk

from plenopticam.gui.constants import PX


# make object for plot widget
class PbarWidget(tk.Frame):

    def __init__(self, parent):

        # inheritance
        tk.Frame.__init__(self, parent)
        self.parent = parent

        # instantiate status
        self.tk_stat = tk.Label(self, text=self.parent.sta.stat_var)
        self.tk_stat.pack(fill='both', expand=True, side='top')

        # style definitions
        self.s = ttk.Style(self)
        self.s.theme_use('alt')
        self.s.layout("LabeledProgressbar",
                 [('LabeledProgressbar.trough',
                   {'children': [('LabeledProgressbar.pbar',
                                  {'side': 'left', 'sticky': 'nswe'}),
                                 ("LabeledProgressbar.label",
                                  {"sticky": ""})],
                  'sticky': 'nswe'})])
        self.s.configure("LabeledProgressbar", foreground='gray1', background='SteelBlue2')    # progress bar color

        # instantiate progress bar
        self.tk_prog_bar = ttk.Progressbar(self, orient=tk.HORIZONTAL, mode='determinate', style="LabeledProgressbar")
        self.tk_prog_bar.pack(fill='both', expand=True, side='bottom', pady=PX, padx=PX)
        self.tk_prog_bar['value'] = 0
        self.tk_prog_bar['maximum'] = 100

        # set-up event tracking
        self.parent.sta.bind_to_stat(self.update_stat_msg)
        self.parent.sta.bind_to_prog(self.update_prog_bar)
        self.parent.sta.bind_to_prog(self.update_prog_msg)

    def update_stat_msg(self, msg):

        self.tk_stat.configure(text=msg)
        self.update()

    def update_prog_bar(self, val):

        if isinstance(val, (float, int)):
            self.tk_prog_bar['value'] = val
        elif isinstance(val, str):
            self.tk_prog_bar['value'] = 0
        self.update()

    def update_prog_msg(self, val):

        if isinstance(val, (float, int)):
            prog_msg = "{0} %       ".format(val)
        else:
            prog_msg = "{0}         ".format(val)

        self.s.configure("LabeledProgressbar", text=prog_msg)
        self.update()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


from plenopticam import misc
from plenopticam.cfg import PlenopticamConfig

# external libs
import numpy as np

try:
    from scipy.signal import medfilt
    from scipy.ndimage import median_filter
except ImportError:
    raise ImportError('Please install scipy package.')

try:
    from colour_demosaicing import demosaicing_CFA_Bayer_bilinear, demosaicing_CFA_Bayer_Malvar2004, demosaicing_CFA_Bayer_Menon2007
except ImportError:
    raise ImportError('Please install colour_demosaicing package')


class CfaProcessor(object):

    def __init__(self, bay_img=None, wht_img=None, cfg=None, sta=None):

        # input variables
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else misc.PlenopticamStatus()
        self._bay_img = bay_img.astype('float32') if isinstance(bay_img, np.ndarray) else None
        self._wht_img = wht_img.astype('float32') if isinstance(wht_img, np.ndarray) else None

        self._bit_pac = self.cfg.lfpimg['bit'] if 'bit' in self.cfg.lfpimg else 10
        self._gains = self.cfg.lfpimg['awb'] if 'awb' in self.cfg.lfpimg else [1, 1, 1, 1]
        self._bay_pat = self.cfg.lfpimg['bay'] if 'bay' in self.cfg.lfpimg else None

        # output variables
        self._rgb_img = np.array([])

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # apply auto white balance gains while considering image highlights
        self.safe_bayer_awb()

        # debayer to rgb image
        if 'bay' in self.cfg.lfpimg.keys() and len(self._bay_img.shape) == 2:
            self.bay2rgb(2)

        # convert to uint16
        self._rgb_img = misc.Normalizer(self._rgb_img).uint16_norm()

        return True

    def bay2rgb(self, method=2):

        # print status
        self.sta.status_msg('Debayering', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # Bayer to RGB conversion
        if method == 0:
            self._rgb_img = demosaicing_CFA_Bayer_bilinear(self._bay_img, self.cfg.lfpimg['bay'])
        elif method == 1:
            self._rgb_img = demosaicing_CFA_Bayer_Malvar2004(self._bay_img, self.cfg.lfpimg['bay'])
        else:
            self._rgb_img = demosaicing_CFA_Bayer_Menon2007(self._bay_img, self.cfg.lfpimg['bay'])

        # normalize image
        min = np.percentile(self._rgb_img, 0.05)
        max = np.max(self.rgb_img)
        self._rgb_img = misc.Normalizer(self._rgb_img, min=min, max=max).type_norm()

        # update status message
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def _reshape_bayer(self):

        if len(self._bay_img.shape) == 2:
            # reshape bayer image to 4 channels in third dimension (G, R, B, G)
            self._bay_img = np.dstack((self._bay_img[0::2, 0::2], self._bay_img[0::2, 1::2],
                                       self._bay_img[1::2, 0::2], self._bay_img[1::2, 1::2]))

        elif len(self._bay_img.shape) == 3:
            if self._bay_img.shape[2] == 4:
                # reshape 4 channel bayer image to 2D bayer image
                arr, self._bay_img = self._bay_img, np.zeros(np.array(self._bay_img.shape[:2])*2)
                self._bay_img[0::2, 0::2] = arr[..., 0]
                self._bay_img[0::2, 1::2] = arr[..., 1]
                self._bay_img[1::2, 0::2] = arr[..., 2]
                self._bay_img[1::2, 1::2] = arr[..., 3]

        return True

    def _reshape_white(self):

        if self._wht_img is not None:
            if len(self._wht_img.shape) == 2:
                # reshape bayer image to 4 channels in third dimension
                self._wht_img = np.dstack((self._wht_img[0::2, 0::2], self._wht_img[0::2, 1::2],
                                           self._wht_img[1::2, 0::2], self._wht_img[1::2, 1::2]))

            elif len(self._wht_img.shape) == 3:
                if self._wht_img.shape[2] == 4:
                    # reshape 4 channel bayer image to 2D bayer image
                    arr, self._wht_img = self._wht_img, np.zeros(np.array(self._wht_img.shape[:2])*2)
                    self._wht_img[0::2, 0::2] = arr[..., 0]
                    self._wht_img[0::2, 1::2] = arr[..., 1]
                    self._wht_img[1::2, 0::2] = arr[..., 2]
                    self._wht_img[1::2, 1::2] = arr[..., 3]

        return True

    @property
    def rgb_img(self):
        return self._rgb_img.copy()

    def apply_awb(self, img_arr=None, bay_pat=None, gains=None):
        """ white balance from gains """

        img_arr = self._bay_img if img_arr is None else img_arr
        bay_pat = self._bay_pat if bay_pat is None else bay_pat
        gains = self._gains if gains is None else gains

        if len(img_arr.shape) == 3:

            img_arr[..., 0] *= gains[0]             # red channel
            img_arr[..., 1] *= gains[1]             # green channel
            img_arr[..., 2] *= gains[2]             # blue channel

        elif len(img_arr.shape) == 2 and bay_pat == "GRBG":

            img_arr[0::2, 0::2] *= gains[0]         # green-red channel
            img_arr[0::2, 1::2] *= gains[1]         # red channel
            img_arr[1::2, 0::2] *= gains[2]         # blue channel
            img_arr[1::2, 1::2] *= gains[3]         # green-blue channel

        elif len(img_arr.shape) == 2 and bay_pat == "BGGR":

            img_arr[0::2, 0::2] *= gains[0]         # blue channel
            img_arr[0::2, 1::2] *= gains[1]         # green-blue channel
            img_arr[1::2, 0::2] *= gains[2]         # green-red channel
            img_arr[1::2, 1::2] *= gains[3]         # red channel

        return img_arr

    def _correct_bayer_highlights(self, gains=None):
        """ inspired by CLIM_VSENSE highlight correction """

        self._gains = self._gains if gains is None else gains

        bay_img = self._bay_img.copy()
        wht_img = np.ones(self._bay_img.shape) if self._wht_img is None else self._wht_img

        fact = self._gains.max()
        ThSatCol = .99

        # Determine the component with lowest value (the last one to saturate).
        min_sat, min_idx = bay_img.min(2), bay_img.argmin(2)
        min_sat_bal = min_sat * self._gains[min_idx]

        # weights
        min_sat = min_sat * wht_img.mean(2)
        min_sat[min_sat > 1] = 1
        min_sat **= 2

        for i in range(4):
            ref_img = np.divide(ThSatCol, wht_img[..., i], out=np.ones_like(wht_img[..., i])*float('Inf'), where=wht_img[..., i]!=0)
            ref_img[~np.isfinite(ref_img)] = 0
            ch_idxs = np.nonzero(bay_img[..., i] > ref_img)
            inm_img = (min_sat_bal*(1-min_sat) + bay_img[..., i]*fact * min_sat) / self._gains[i]
            bay_img[..., i][ch_idxs] = np.maximum.reduce([bay_img[..., i][ch_idxs], inm_img[ch_idxs]])

        self._bay_img = bay_img

        return True

    @staticmethod
    def desaturate_clipped(img_arr, gains=None):

        # skip process if gains not set
        if gains is None:
            return img_arr

        # original channel intensities
        orig = img_arr / gains

        # identify clipped pixels
        beta = orig / np.amax(orig, axis=2)[..., np.newaxis]
        weights = beta * gains
        weights[weights < 1] = 1
        mask = np.zeros(orig.shape[:2])
        mask[np.amax(orig, axis=2) >= orig.max()] = 1

        # de-saturate clipped values
        img_arr[mask > 0] /= weights[mask > 0]

        return img_arr

    @staticmethod
    def correct_color(img, ccm_mat=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])):
        """ color correction according to http://www.imatest.com/docs/colormatrix/ using Case 1 """

        # perform color correction
        img_ccm = np.dot(np.vstack(img), ccm_mat).reshape(img.shape)

        return img_ccm

    def safe_bayer_awb(self):

        self._reshape_bayer()
        self._reshape_white()

        self.set_gains(self.cfg.lfpimg['awb'])
        self._correct_bayer_highlights()

        self._reshape_bayer()
        self._reshape_white()

        self.apply_awb()
        self._bay_img[self._bay_img < 0] = 0

        if 'exp' in self.cfg.lfpimg:
            exp_bias = self.cfg.lfpimg['exp']
            max_lum = 2**(-exp_bias)
            self._bay_img = self.soft_clipping(self._bay_img/max_lum, 7)*max_lum

    def soft_clipping(self, img, max):

        b = np.exp(max)
        return np.log((1+b)/(1+b*np.exp(-max*img)))/np.log(1+b)

    def set_gains(self, gains=None):

        # skip process if gains not set
        if gains is not None:
            if len(self._bay_img.shape) == 2 or len(self._bay_img.shape) == 3 and self._bay_img.shape[-1] == 4:
                if self._bay_pat == "GRBG":
                    self._gains = np.array([gains[2], gains[1], gains[0], gains[3]])
                elif self._bay_pat == "BGGR":
                    self._gains = np.array([gains[0], gains[2], gains[3], gains[1]])
                else:
                    return None
            elif len(self._bay_img.shape) == 3 and self._bay_img.shape[-1] == 3:
                self._gains = np.array([gains[1], (gains[2] + gains[3]) / 2, gains[0]])
            else:
                return None
        else:
            return None

        return self._gains



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import scipy

from plenopticam.lfp_calibrator.centroid_drawer import CentroidDrawer
from plenopticam.misc import PlenopticamStatus
from plenopticam.cfg import PlenopticamConfig


class LfpRotator(object):

    def __init__(self, lfp_img, mic_list, rad=None, cfg=None, sta=None):

        # input and output variable
        self._lfp_img = lfp_img
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()

        # internal variables
        self._centroids = np.asarray(mic_list)
        self._rad = rad

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # determine rotation angle (radians) of micro image centers (skip if angle already set)
        self._estimate_rad() if self._rad is None else None

        # rotate image
        self._rotate_img()

        # rotate centroids
        self._rotate_centroids()

        # write plot img to hard drive (debug only)
        if self.cfg.params[self.cfg.opt_dbug] and not self.sta.interrupt:
            self.sta.status_msg('Save rotated image (debug mode)')
            self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])
            CentroidDrawer(self._lfp_img, self._centroids, self.cfg).write_centroids_img(fn='lfp_rotated.png')
            self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def _estimate_rad(self, regression=True):

        # status update
        self.sta.status_msg('Rotation angle estimation', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # middle row and column indices
        self._mid_row_idx = int(self._centroids[:, 2].max()/2)
        self._mid_col_idx = int(self._centroids[:, 3].max()/2)

        # get central row and column
        mid_row = self._centroids[self._centroids[:, 2] == self._mid_row_idx][:, :2]
        mid_col = self._centroids[self._centroids[:, 3] == self._mid_col_idx][:, :2]
        mid_col = mid_col[0::2]     # leave out every other centroid in column to compensate for hexagonal array

        if regression:
            # ordinary least-squares linear regression
            slope_row, slope_col = self._regress_method(mid_row, mid_col)
        else:
            # mean gradient (for validation)
            slope_row, slope_col = self._gradient_method(mid_row, mid_col)

        # convert to radians
        rad_row = np.arctan(slope_row)
        rad_col = np.arctan(slope_col)

        # set rotation angle by averaging along both dimensions
        self._rad = (rad_row+rad_col*-1)/2

        # status update
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    @staticmethod
    def _regress_method(mid_row, mid_col):
        """ get slopes via ordinary least-squares linear regression """

        # horizontal slope
        A = np.vstack([np.ones(len(mid_row[:, 1])), mid_row[:, 1]]).T
        b = mid_row[:, 0]
        x = np.dot(np.linalg.pinv(A), b)
        slope_row = x[1]

        # vertical slope
        A = np.vstack([np.ones(len(mid_col[:, 0])), mid_col[:, 0]]).T
        b = mid_col[:, 1]
        x = np.dot(np.linalg.pinv(A), b)
        slope_col = x[1]

        return slope_row, slope_col

    @staticmethod
    def _gradient_method(mid_row, mid_col):
        """ get slopes for rotation using mean gradient """

        slope_row = np.mean(np.diff(mid_row[:, 0])/np.diff(mid_row[:, 1]))     # mean(diff(y)/diff(x))
        slope_col = np.mean(np.diff(mid_col[:, 1])/np.diff(mid_col[:, 0]))

        return slope_row, slope_col

    def _rotate_img(self):

        # status update
        self.sta.status_msg('Rotate image', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # convert radians to degrees
        deg_angle = self._rad*180/np.pi

        if deg_angle != 0:
            # rotate images (counter-clockwise) without reshape to retain shape of array
            self._lfp_img = scipy.ndimage.rotate(self._lfp_img, angle=deg_angle, reshape=False, order=3)

        # status update
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def _rotate_centroids(self):
        """ transformation of centroids via translation and rotation """

        # status update
        self.sta.status_msg('Rotate centroids', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # get image center coordinates
        img_center = (np.asarray(self._lfp_img.shape[:2])-1)/2.  # -1 for .5 decrement compensation in scipy rotate

        # translate data points to origin
        self._centroids[:, :2] -= img_center

        # matrix for counter-clockwise rotation around z-axis
        Rz = np.array([[np.cos(self._rad), -np.sin(self._rad)], [np.sin(self._rad), np.cos(self._rad)]])

        # rotate data points around z-axis
        self._centroids[:, :2] = np.dot(Rz, self._centroids[:, :2].T).T

        # translate data points back to image center
        self._centroids[:, :2] += img_center

        # status update
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    @property
    def lfp_img(self):
        return self._lfp_img.copy()

    @property
    def centroids(self):
        return self._centroids.tolist()

    def _rota_plot(self, slope):
        """ plots for debug purposes only """

        import matplotlib.pyplot as plt
        plt.figure()
        plt.imshow(self._lfp_img)
        plt.plot(np.array(self._centroids)[:, 1], np.array(self._centroids)[:, 0], 'rx')

        central_centroid = self._centroids[(self._centroids[:, 2] == self._mid_row_idx) &
                                           (self._centroids[:, 3] == self._mid_col_idx)].ravel()

        p1 = central_centroid[:2]
        p2 = ((self._lfp_img.shape[1]-central_centroid[1])*slope+central_centroid[0], self._lfp_img.shape[1])

        plt.plot((p1[1], p2[1]), (p1[0], p2[0]), 'b-')

        p3 = (self._lfp_img.shape[0], (self._lfp_img.shape[0]-central_centroid[0])*slope+central_centroid[1])

        plt.plot((p1[1], p3[1]), (p1[0], p3[0]), 'b-')



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam import misc
from plenopticam.lfp_extractor.lfp_viewpoints import LfpViewpoints


# external libs
import os
import numpy as np
from color_space_converter import rgb2gry


class LfpExporter(LfpViewpoints):

    def __init__(self, refo_stack=None, *args, **kwargs):
        super(LfpExporter, self).__init__(*args, **kwargs)

        self.refo_stack = np.asarray(refo_stack) if refo_stack is not None else None

    def write_viewpoint_data(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # write central view as thumbnail image
        self.export_thumbnail(type='png')

        # write viewpoint image files to hard drive
        self.export_viewpoints(type='png')

        # write viewpoints as single image
        if self.cfg.params[self.cfg.opt_dbug]:
            self.export_vp_stack(type='png')

        # write viewpoint gif animation
        self.gif_vp_img(duration=.1)

        return True

    def export_thumbnail(self, type='tiff'):

        thumb = misc.Normalizer(self.central_view.copy()).type_norm()

        # export central viewpoint as thumbnail image
        fp = os.path.join(self.cfg.exp_path, 'thumbnail')
        misc.save_img_file(thumb, file_path=fp, file_type=type, tag=self.cfg.params[self.cfg.opt_dbug])

        return True

    def export_viewpoints(self, type='tiff'):

        # print status
        self.sta.status_msg('Write viewpoint images', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        ptc_leng = self.cfg.params[self.cfg.ptc_leng]

        # create folder
        folderpath = os.path.join(self.cfg.exp_path, 'viewpoints_'+str(ptc_leng)+'px')
        misc.mkdir_p(folderpath)

        # normalize image array to 16-bit unsigned integer
        vp_img_arr = misc.Normalizer(self.vp_img_arr).uint16_norm()

        # export viewpoint images as image files
        for j in range(ptc_leng):
            for i in range(ptc_leng):

                misc.save_img_file(vp_img_arr[j, i], os.path.join(folderpath, str(j).zfill(2) + '_' + str(i).zfill(2)),
                                   file_type=type, tag=self.cfg.params[self.cfg.opt_dbug])

                # print status
                percentage = (((j * self._size_pitch + i + 1) / self._size_pitch ** 2) * 100)
                self.sta.progress(percentage, self.cfg.params[self.cfg.opt_prnt])

                if self.sta.interrupt:
                    return False

        return True

    def export_vp_stack(self, type='png', downscale=None):
        """ write viewpoint images stitched to together in a single image """

        # print status
        self.sta.status_msg('Write viewpoint image stack', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # downscale image
        downscale = True if downscale is None else downscale
        views_stacked_img = misc.img_resize(self.views_stacked_img.copy(), 1 / self._size_pitch) \
            if downscale else self.views_stacked_img.copy()

        # normalization
        p_lo = np.percentile(rgb2gry(self.central_view), 0.05)
        p_hi = np.percentile(rgb2gry(self.central_view), 99.995)
        views_stacked_img = misc.Normalizer(views_stacked_img, min=p_lo, max=p_hi).uint8_norm()

        # export all viewpoints in single image
        views_stacked_path = os.path.join(self.cfg.exp_path, 'views_stacked_img_' + str(self._size_pitch) + 'px')
        misc.save_img_file(views_stacked_img, file_path=views_stacked_path, file_type=type)

        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def export_refo_stack(self, file_type=None):

        # print status
        self.sta.status_msg('Write refocused images', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        refo_stack = misc.Normalizer(self.refo_stack).uint16_norm()
        if self.cfg.params[self.cfg.opt_refi]:
            a_list = np.arange(*np.array(self.cfg.params[self.cfg.ran_refo]) * self.cfg.params[self.cfg.ptc_leng])
            a_list = np.round(a_list / self.cfg.params[self.cfg.ptc_leng], 2)
        else:
            a_list = range(*self.cfg.params[self.cfg.ran_refo])

        # create folder
        string = 'subpixel_' if self.cfg.params[self.cfg.opt_refi] else ''
        folder_path = os.path.join(self.cfg.exp_path, 'refo_' + string + str(self._size_pitch) + 'px')
        misc.mkdir_p(folder_path)

        for i, refo_img in enumerate(refo_stack):

            # get depth plane number for filename
            a = a_list[i]

            self.save_refo_slice(a, refo_img, folder_path, file_type=file_type)

            # print status
            percentage = ((i+1) / len(refo_stack)) * 100
            self.sta.progress(percentage, self.cfg.params[self.cfg.opt_prnt])

        return True

    def save_refo_slice(self, a, refo_img, folder_path=None, file_type=None, string=None):

        string = 'subpixel_' if self.cfg.params[self.cfg.opt_refi] and string is None else string

        if folder_path is None:
            folder_path = os.path.join(self.cfg.exp_path, 'refo_' + string + str(self._size_pitch) + 'px')
            misc.mkdir_p(folder_path)

        # write image file
        misc.save_img_file(refo_img, os.path.join(folder_path, str(a)), file_type=file_type)

        return True

    def gif_vp_img(self, duration, pattern='circle'):

        # micro image size selection
        try:
            mi_size = np.min(self.cfg.calibs[self.cfg.ptc_mean])
        except (KeyError, TypeError):
            mi_size = self.cfg.params[self.cfg.ptc_leng]

        # filter images forming a pattern
        lf_radius = min(int((mi_size+1)//4), self._cent_pitch)
        img_set = self.reorder_vp_arr(pattern=pattern, lf_radius=lf_radius)

        # image normalization
        img_set = misc.Normalizer(img_set).uint8_norm()

        # export gif animation
        fn = 'view_animation_' + str(lf_radius * 2 + 1) + 'px'
        misc.save_gif(img_set, duration=duration, fp=self.cfg.exp_path, fn=fn)

        return True

    def gif_refo(self):

        # image normalization
        refo_stack = misc.Normalizer(self.refo_stack).uint8_norm()

        # append reversed array copy to play forward and backwards
        refo_stack = np.concatenate((refo_stack, refo_stack[::-1]), axis=0)

        # export gif animation
        fn = 'refocus_animation_' + str(self.cfg.params[self.cfg.ptc_leng]) + 'px'
        misc.save_gif(refo_stack, duration=.5, fp=self.cfg.exp_path, fn=fn)

        return True



# dictionary keys for configuration parameters in json file
PARAMS_KEYS = (
    # strings
    'lfp_path',
    'cal_path',
    'cal_meta',
    'cal_meth',
    'smp_meth',
    # integers
    'ptc_leng',
    # lists
    'ran_refo',
    # booleans
    'opt_cali',
    'opt_vign',
    'opt_lier',
    'opt_cont',
    'opt_colo',
    'opt_awb_',
    'opt_sat_',
    'opt_view',
    'opt_refo',
    'opt_refi',
    'opt_pflu',
    'opt_arti',
    'opt_rota',
    'opt_dbug',
    'opt_prnt',
    'opt_dpth',
    'dir_remo'
)

# dictionary values for configuration parameters in json file
PARAMS_VALS = (
    # strings
    '',
    '',
    '',
    '',
    '',
    # integers
    7,
    # lists
    [0, 2],
    # booleans
    False,
    True,
    False,
    False,
    True,
    False,
    False,
    True,
    True,
    False,
    False,
    False,
    False,
    False,
    True,
    True,
    False
)

PARAMS_TYPE = (
    # strings
    'str',
    'str',
    'str',
    'sel',
    'sel',
    # integers
    'sel',
    # lists
    'ran',
    # booleans
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool',
    'bool'
)

PARAMS_NAME = (
    # strings
    'Light field image',
    'Calibration source',
    'Metadata file',
    'Calibration method',
    'Resampling method',
    # integers
    'Micro image patch size',
    # lists
    'Refocusing range',
    # booleans
    'Redo calibration',
    'De-Vignetting',
    'Pixel outlier removal',
    'Auto-Contrast',
    'Color equalization',
    'White balance',
    'Auto-Saturation',
    'Viewpoint synthesis',
    'Refocus synthesis',
    'Refocus refinement',
    'Scheimpflug focus',
    'Hex-Artifact removal',
    'Rotation of light field',
    'Debug option',
    'Status print option',
    'Depth map',
    'Remove output folder'
)

# dictionary keys for calibration parameters names in json file
CALIBS_KEYS = (
   # calibs dict keys
   'pat_type',
   'ptc_mean',
   'mic_list'
)

# value ranges
PFLU_VALS = ('vertical', 'horizontal', 'skew up', 'skew down')
PTCH_SIZE = list(range(3, 99, 2))
CALI_METH = ('area', 'peak', 'grid-fit', 'vign-fit', 'corn-fit')
SMPL_METH = ('global', 'local')

# command line interface options
CLIF_SHRT = "ghf:c:p:r:m:s:"
CLIF_OPTS = [
    # non-setting commands
    "gui", "help",
    # strings
    "file=",
    "cali=",
    "meta=",
    "meth=",
    "smpl=",
    # integers
    "patch=",
    # lists
    "refo=",
    # booleans
    "rcal",
    "vign",
    "lier",
    "cont",
    "colo",
    "awb_",
    "sat_",
    "view",
    "ropt",
    "refi",
    "pflu",
    "arti",
    "rota",
    "dbug",
    "prnt",
    "dpth",
    "remo"
]



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from setuptools import setup, find_packages
from plenopticam import __version__
from sys import platform
from docutils import core
import os

APP = ['plenopticam/gui/top_level.py']

FILES = [
        # ('subdir' , ['file_path'])
        ('cfg', ['plenopticam/cfg/cfg.json']),
        ('gui/icns', ['plenopticam/gui/icns/1055104.icns']),
        ('gui/icns', ['plenopticam/gui/icns/1055104.ico']),
        ('gui/icns', ['plenopticam/gui/icns/1055104.gif'])
]

OPTIONS = {
    "argv_emulation": True,
    "compressed": True,
    "optimize": 2,
    "iconfile": 'plenopticam/gui/icns/1055104.icns',
    "excludes": ['matplotlib'],
    "plist": dict(NSHumanReadableCopyright='2020 Christopher Hahne'),
    "packages": ['numpy', 'scipy', 'imageio', 'docutils', 'PIL', 
                 'colour-demosaicing', 'colour', 'color-matcher', 'color-space-converter'],
}

if platform == 'darwin':
    extra_options = dict(
        setup_requires=['py2app'],
        app=APP,
        data_files=FILES,
        options=dict(py2app=OPTIONS),
    )
elif platform == 'win32':
    extra_options = dict(
        setup_requires=[],
        #app=APP,
        data_files=FILES,
    )
else:
    extra_options = dict(
        setup_requires=[],
        data_files=FILES,
 )

path = os.path.dirname(os.path.realpath(__file__))
# parse description section text
readme_path = os.path.join(path, 'README.rst')
with open(readme_path, 'r') as f:
    data = f.read()
    readme_nodes = list(core.publish_doctree(data))
    for node in readme_nodes:
        if node.astext().startswith('Description'):
                long_description = node.astext().rsplit('\n\n')[1]

# parse package requirements from text file
reqtxt_path = os.path.join(path, 'requirements.txt')
with open(reqtxt_path, 'r') as f:
    req_list = f.read().split('\n')

setup(
      name='plenopticam',
      version=__version__,
      description='Light field photography application',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      url='http://github.com/hahnec/plenopticam',
      author='Christopher Hahne',
      author_email='inbox@christopherhahne.de',
      license='GNU GPL V3.0',
      keywords='lightfield plenoptic rendering engine image processing software application lytro toolbox calibration '
               'light field depth refocusing refocus baseline disparity resolution',
      scripts=['plenopticam/bin/cli_script.py'],
      entry_points={'console_scripts': ['plenopticam=plenopticam.bin.cli_script:main'], },
      packages=find_packages(),
      install_requires=req_list,
      include_package_data=True,
      zip_safe=False,
      **extra_options
      )



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2021 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
from scipy.spatial.distance import cdist

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.lfp_calibrator.centroid_sorter import CentroidSorter
from plenopticam.lfp_calibrator.grid_fitter import GridFitter


class CentroidFitSorter(CentroidSorter, GridFitter):

    # force hexagonal shift of second row to be on the right of the upper left
    #_hex_odd = True

    def __init__(self, *args, **kwargs):
        super(CentroidFitSorter, self).__init__(*args, **kwargs)

        self.corner_mics = []

        self.keep_pts_opt = False if 'keep_pts_opt' not in kwargs else kwargs['keep_pts_opt']

    def corner_fit(self, norm_type: int = 1):

        # estimate MLA dimensions
        self._mla_dims(counterclockwise_opt=False)
        self._get_mla_pitch()
        self._estimate_mla_geometry(self._pitch[0])

        # extract MLA corner centroids
        self.corner_mics = np.vstack([self._upper_l, self._upper_r, self._lower_l, self._lower_r])
        self.corner_mics = np.hstack([self.corner_mics, np.array([[0, 0], [0, 1], [1, 0], [1, 1]])])

        # 4-corners (8 coordinate) projection fit
        corner_fitter = GridFitter(coords_list=self.corner_mics, pat_type=self.pattern, hex_odd=None, normalize=norm_type)
        corner_fitter._MAX_X = self._lens_x_max
        corner_fitter._MAX_Y = self._lens_y_max
        corner_fitter.coeff_fit(euclid_opt=False)
        coeffs = corner_fitter._coeffs

        # generate grid based on 4-corner (8 coordinate) fit
        ruff_pts = GridFitter.grid_gen(dims=[self._lens_y_max, self._lens_x_max], pat_type=self.pattern, hex_odd=None, normalize=norm_type)
        ruff_fit = corner_fitter.apply_transform(coeffs, ruff_pts.copy())
        del corner_fitter

        # validation
        for pts in [self._centroids, ruff_fit]:
            dres = cdist(pts[:, :2], self.corner_mics[:, :2], 'euclidean')
            self.idxs = np.argmin(dres, axis=0)
            assert np.sum(np.square(pts[self.idxs, :2] - self.corner_mics[:, :2])) < min(self._pitch)/2

        # validate centroid index assignments
        assert ruff_fit[self.idxs][0][2] == 0 and ruff_fit[self.idxs][0][3] == 0
        assert ruff_fit[self.idxs][1][2] == 0 and ruff_fit[self.idxs][1][3] == max(ruff_fit[:, 3])
        assert ruff_fit[self.idxs][2][2] == max(ruff_fit[:, 2]) and ruff_fit[self.idxs][2][3] == 0
        assert ruff_fit[self.idxs][3][2] == max(ruff_fit[:, 2]) and ruff_fit[self.idxs][3][3] == max(ruff_fit[:, 3])

        return ruff_fit.tolist()

    def match_points(self, centroids, fit_points):

        sorted_mics = fit_points.copy()
        centroid_rows = int(max(fit_points[:, 2]))

        # iterate through centroid rows (cdist array can easily require large amounts of memory)
        for i in range(centroid_rows):

            # select next row of indexed points
            frow = fit_points[fit_points[:, 2] == i]

            # measure distance between fitted and actual centroids
            dres = cdist(centroids[:, :2], frow[:, :2], 'euclidean')

            # determine indices and distance of closest actual centroids to each fitted/index point
            idxs = np.argmin(dres, axis=0)
            mins = np.min(dres, axis=0)

            # use threshold to exclude points which are too far away
            thrs = mins < min(self._pitch)/2

            # copy actual centroid peaks (fulfilling euclidean threshold) to the fitted grid
            sorted_mics[sorted_mics[:, 2] == i][thrs, :2] = self._centroids[:, :2][idxs][thrs, :2]

        return sorted_mics

    def main(self, keep_pts_opt=None):

        self.keep_pts_opt = keep_pts_opt if keep_pts_opt is not None else self.keep_pts_opt

        ruff_fit = self.corner_fit()
        if self.keep_pts_opt:
            self._mic_list = self.match_points(self._centroids, np.array(ruff_fit))
        else:
            self._mic_list = ruff_fit

        # attach results to config object if present
        if hasattr(self, 'cfg') and hasattr(self.cfg, 'calibs'):
            self.cfg.calibs[self.cfg.mic_list] = self._mic_list
            self.cfg.calibs[self.cfg.pat_type] = self._pattern
            self.cfg.calibs[self.cfg.ptc_mean] = self._pitch

        return ruff_fit